#import tadashi
from tadashi import TrEnum

from app import miniAMR

app = miniAMR(run_args=["--nx", "40", "--ny", "40", "--nz", "40"])


app.compile()
original_time = app.measure()
print(f"{original_time=}")
app.reset_scops()
trs = [
        [0, 6, TrEnum.FULL_SPLIT],
        #[0, 3, TrEnum.TILE2D, tile_size, tile_size],
        #[0, 9, TrEnum.TILE3D, tile_size, tile_size, tile_size],
    ]
app.transform_list(trs)
new_app = app.generate_code(ephemeral=False)
new_app.compile()
new_time = new_app.measure()
print(f"{new_time=}")
speedup = original_time /new_time
print(f"{speedup=}")

print("DONE")



#!/bin/env python
from pathlib import Path
from random import choice, seed

import tadashi
from tadashi.apps import Simple

seed(1234)

dir_path = Path(__file__).parent
examples_path = dir_path if dir_path.name == "examples" else "examples"
app = Simple(f"{examples_path}/inputs/depnodep.c")
print(app)

node = app.scops[0].schedule_tree[1]
print(f"{node=}")
tr = choice(node.available_transformations)
print(f"{tr=}")
# output:

args = choice(node.get_args(tr, -10, 10))
print(f"{args=}")
# output:

legal = node.transform(tr, *args)
print(f"{legal=}")
# output:

app.compile()
print(f"{app.measure()=}")
transformed_app = app.generate_code()
transformed_app.compile()
print(f"{transformed_app.measure()=}")
# output:



#!/usr/bin/env python3
import re
import sys

C_FILE = "tadashi/src/transformations.c"
H_FILE = "tadashi/src/transformations.h"
PXD_FILE = "tadashi/transformations.pxd"
PY_FILE = "tadashi/_tr_wrappers.py"

HEADER_H = """/** @file */
#ifndef TRANSFORMATIONS_H
#define TRANSFORMATIONS_H

#include <isl/schedule.h>
#include <isl/schedule_node.h>

#if defined(__cplusplus)
extern "C" {
#endif

"""

FOOTER_H = """
#if defined(__cplusplus)
}
#endif

#endif // TRANSFORMATIONS_H
"""

HEADER_PXD = """# -*- mode:cython -*-

from tadashi.isl cimport *

cdef extern from "transformations.h":
"""

HEADER_PY = """#!/bin/env python

import cython
from cython.cimports.tadashi import isl
from cython.cimports.tadashi.scop import Scop
from cython.cimports.tadashi.transformations import *

"""

# Regex to find tadashi transformations
# Look for: __isl_give isl_schedule_node * tadashi_name(...) {
# This is a bit brittle but should work for the coding style used.
REGEX = re.compile(
    r"^(__isl_give\s+)?isl_schedule_node\s*\*\s*\n?" r"(tadashi_\w+)\(([^)]+)\)\s*\{",
    re.MULTILINE | re.DOTALL,
)


def parse_args(args_str):
    """Parse C arguments into a list of (type, name)."""
    # Remove newlines and extra spaces
    args_str = " ".join(args_str.split())
    args_list = []

    # Split by comma
    parts = args_str.split(",")
    for part in parts:
        part = part.strip()
        if not part:
            continue

        # Simple heuristic: last word is name, rest is type
        # Handle pointers and annotations
        part = (
            part.replace("__isl_take ", "")
            .replace("__isl_keep ", "")
            .replace("__isl_give ", "")
        )

        # Check for pointer *
        if "*" in part:
            type_part, name_part = part.rsplit("*", 1)
            type_part += "*"
            name_part = name_part.strip()
            if not name_part:  # e.g. "int *"
                continue
        else:
            # splitting "int size"
            subparts = part.rsplit(" ", 1)
            if len(subparts) == 2:
                type_part, name_part = subparts
            else:
                # Fallback or weird case
                type_part = part
                name_part = "arg"

        args_list.append((type_part.strip(), name_part.strip()))
    return args_list


def generate():
    with open(C_FILE, "r") as f:
        content = f.read()

    funcs = []
    for match in REGEX.finditer(content):
        func_name = match.group(2)
        args_str = match.group(3)
        args = parse_args(args_str)
        funcs.append((func_name, args))

    # Generate .h
    with open(H_FILE, "w") as f:
        f.write(HEADER_H)
        for name, args in funcs:
            arg_str = ", ".join([f"{t} {n}" for t, n in args])
            f.write(f"isl_schedule_node *{name}({arg_str});\n")
        f.write(FOOTER_H)

    # Generate .pxd
    with open(PXD_FILE, "w") as f:
        f.write(HEADER_PXD)
        for name, args in funcs:
            arg_str = ", ".join([f"{t} {n}" for t, n in args])
            f.write(f"    cdef isl_schedule_node *{name}({arg_str})\n")

    # Generate .py
    with open(PY_FILE, "w") as f:
        f.write(HEADER_PY)
        for name, args in funcs:
            py_name = name.replace("tadashi_", "")

            # Build python signature and C call args
            py_args = []
            c_args = []

            # First arg is always node/scop
            # Check if first arg is isl_schedule_node*
            if args and "isl_schedule_node" in args[0][0]:
                py_args.append("scop: Scop")
                c_args.append("scop.ptr_ccscop.current_node")
                start_idx = 1
            else:
                # Should not happen for these transformations but safer
                start_idx = 0

            for t, n in args[start_idx:]:
                # Map C types to Cython/Python types
                if t == "long":
                    py_type = "int"
                else:
                    py_type = t  # Fallback

                py_args.append(f"{n}: {py_type}")
                c_args.append(n)

            py_sig = ", ".join(py_args)
            c_call_args = ", ".join(c_args)

            f.write(f"@cython.ccall\n")
            f.write(f"def {py_name}({py_sig}):\n")
            f.write(f"    scop.ptr_ccscop.current_node = {name}({c_call_args})\n\n")

    print(f"Generated {H_FILE}, {PXD_FILE}, and {PY_FILE} with {len(funcs)} functions.")


if __name__ == "__main__":
    generate()



#!/bin/env python
import argparse
import json
from pathlib import Path

import matplotlib as mpl
import matplotlib.gridspec as grid_spec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.lines import Line2D

plt.style.use("seaborn-v0_8-paper")
sns.set_theme(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})

plt.rcParams["text.latex.preamble"] = (
    # r"\usepackage{libertine}\usepackage{zi4}\usepackage{newtxmath}"
    r"\usepackage{newtxtext,bm}\usepackage[cmintegrals]{newtxmath}"
)

# begin SC subbmision code
params = {
    "axes.titlesize": 16,
    "axes.labelsize": 16,
    "legend.fontsize": 16,
    "text.usetex": True,
    "font.size": 11,
    "font.family": "libertine",
    # "text.latex.unicode": True,
}
plt.rcParams.update(params)
# end SC subbmision code


def parse_jsonl(filepath):
    cnt_evals = []
    speedups = []
    with open(filepath, "r") as f:
        for line in f:
            # Trim whitespace and check if line is empty
            line = line.strip()
            if not line:
                continue
            data = json.loads(line)
            # Check if required keys exist
            if "cnt_evals" in data and "speedup" in data:
                cnt_evals.append(data["cnt_evals"])
                speedups.append(data["speedup"])
            else:
                print(
                    f"Warning: Skipping line in {filepath} due to missing keys: {line}"
                )
    # Ensure lists are sorted by cnt_evals for proper plotting
    return cnt_evals, speedups


class AnchoredScaleBar(mpl.offsetbox.AnchoredOffsetbox):
    """size: length of bar in data units
    extent : height of bar ends in axes units"""

    def __init__(
        self,
        size=1,
        extent=0.02,
        label="",
        loc=2,
        ax=None,
        pad=0.4,
        borderpad=0.5,
        ppad=0,
        sep=2,
        prop=None,
        frameon=True,
        linekw={},
        **kwargs,
    ):
        if not ax:
            ax = plt.gca()
        trans = ax.get_yaxis_transform()
        size_bar = mpl.offsetbox.AuxTransformBox(trans)
        line = Line2D([0, 0], [0, size], **linekw)
        vline1 = Line2D([-extent / 2.0, extent / 2.0], [0, 0], **linekw)
        vline2 = Line2D([-extent / 2.0, extent / 2.0], [size, size], **linekw)
        size_bar.add_artist(line)
        size_bar.add_artist(vline1)
        size_bar.add_artist(vline2)
        txt = mpl.offsetbox.TextArea(
            label,
            # minimumdescent=False
        )
        self.vpac = mpl.offsetbox.VPacker(
            children=[size_bar, txt], align="center", pad=ppad, sep=sep
        )
        mpl.offsetbox.AnchoredOffsetbox.__init__(
            self,
            loc,
            pad=pad,
            borderpad=borderpad,
            child=self.vpac,
            prop=prop,
            frameon=frameon,
            **kwargs,
        )


def to_csv(json_files, out_path):
    final_speedups = {}
    for f in json_files:
        _, speedups = parse_jsonl(f)
        kernel = f.name.split(".")[0]
        final_speedups[kernel] = speedups[-1]
    df = pd.Series(final_speedups)
    df.rename("MCTS", inplace=True)
    df.index.name = "benchmark"
    df.to_csv(out_path)


def main(datadir, scale_factor):
    fig = plt.figure(figsize=(5, 5))
    i = 0
    ax_objs = []
    directory = Path(datadir)
    json_files = list(sorted(directory.glob("*.jsonl")))
    to_csv(json_files, directory.with_suffix(".csv"))
    filter = [
        "3mm",
        "adi",
        "atax",
        "bicg",
        "covariance",
        "deriche",
        "durbin",
        "fdtd-2d",
        "floyd-warshall",
        "gemver",
        "gesummv",
        "gramschmidt",
        "jacobi-1d",
        "jacobi-2d",
        "lu",
        "ludcmp",
        "mvt",
        "symm",
        "trisolv",
        "trmm",
    ]
    json_files = [f for f in json_files if f.name.split(".")[0] not in filter]

    # cmap = plt.colormaps['brg']
    cmap = plt.colormaps["winter"]

    # cmap = sns.color_palette("blend:#21599c,#b59610", as_cmap=True)
    # cmap = sns.color_palette("blend:#7AB,#EDA", as_cmap=True)
    colors = [cmap(i) for i in np.linspace(0, 1, len(json_files) + 3)]  # Evenly spaced
    # colors = sns.color_palette('tab20', n_colors=len(json_files))
    # colors = sns.cubehelix_palette(len(json_files), rot=-.25, light=.7)

    gs = grid_spec.GridSpec(len(json_files), 1)
    for f in json_files:
        cnt_evals, speedups = parse_jsonl(f)
        ax_objs.append(fig.add_subplot(gs[i : i + 1, 0:]))
        ax = ax_objs[-1]
        # ax.plot(cnt_evals, speedups,
        #         drawstyle="steps-post",
        #         color="#000000")
        scaled = [s * scale_factor for s in speedups]
        ax.fill_between(
            cnt_evals,
            scaled,
            alpha=0.9,
            step="post",
            color="#74a892",  # colors was here
        )
        ax.set_xlim(0, 2000)
        ax.set_ylim(0, 100)

        # make background transparent
        rect = ax_objs[-1].patch
        rect.set_alpha(0)

        # remove borders, axis ticks, and labels
        ax_objs[-1].set_yticklabels([])

        if i != len(json_files) - 1:
            #     ax_objs[-1].set_xlabel(
            #         "Number of Evaluations",
            #         fontsize=10,
            #         # fontweight="bold"
            #     )
            # else:
            ax_objs[-1].set_xticklabels([])

        # spines = ["top","right","left","bottom"]
        # ax.spines[['right', 'top', 'left', 'bottom']].
        spines = ["top", "right", "left"]
        for s in spines:
            ax.spines[s].set_visible(False)
        # ax.spines["bottom"].set_linestyle(":")
        ax.spines["bottom"].set_linestyle((0, (4, 4)))
        ax.spines["bottom"].set_linewidth(0.5)
        kernel = f.name.split(".")[0]
        # print(f)
        ax_objs[-1].text(
            -20,
            0,
            kernel + f"\n({speedups[-1]:0.1F}x)",
            fontsize=10,
            ha="right",
            # color=colors[i]
        )
        i += 1

    gs.update(hspace=-0.57)
    ob = AnchoredScaleBar(
        size=500 * scale_factor,
        label="500x",
        loc=4,
        frameon=True,
        pad=0.6,
        sep=4,
        linekw=dict(color="black", linewidth=0.5),
    )
    ax.add_artist(ob)
    # fig.text(0.07,0.85,"Distribution of Aptitude Test Results from 18 – 24 year-olds",fontsize=20)

    plt.tight_layout()
    # plt.show()
    plt.savefig(directory.with_suffix(".pdf"), bbox_inches="tight")

    # sns.color_palette("blend:#21599c,#b59610", as_cmap=True)
    # sns.color_palette("dark:#5A9_r", as_cmap=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("datadir")
    parser.add_argument("--scale-factor", type=float, default=0.08)
    args = parser.parse_args()
    main(args.datadir, args.scale_factor)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import os
import re
import sys

parent = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, parent)

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Tadashi"
copyright = "2024"
author = "Vatai et al"
release = re.sub("^v", "", os.popen("git describe --tags").read().strip())


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.githubpages",
    "sphinx.ext.ifconfig",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",  # did't work when I tried :(
    "sphinxcontrib.katex",
    "sphinxcontrib.mermaid",
    "sphinx_rtd_theme",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
autoclass_content = "both"
default_role = "py:obj"
todo_include_todos = True
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}
autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# html_theme = "alabaster"
html_theme = "sphinx_rtd_theme"
# html_static_path = ["_static"]



# -*- mode:cython -*-

from tadashi.isl cimport *

cdef extern from "transformations.h":
    cdef isl_schedule_node *tadashi_tile_1d(isl_schedule_node * node, int tile_size)
    cdef isl_schedule_node *tadashi_tile_2d(isl_schedule_node * node, int size1, int size2)
    cdef isl_schedule_node *tadashi_tile_3d(isl_schedule_node * node, int size1, int size2, int size3)
    cdef isl_schedule_node *tadashi_interchange(isl_schedule_node * node)
    cdef isl_schedule_node *tadashi_full_fuse(isl_schedule_node * node)
    cdef isl_schedule_node *tadashi_fuse(isl_schedule_node * node, int idx1, int idx2)
    cdef isl_schedule_node *tadashi_full_split(isl_schedule_node * node)
    cdef isl_schedule_node *tadashi_split(isl_schedule_node * node, int split_idx)
    cdef isl_schedule_node *tadashi_scale(isl_schedule_node * node, long scale)
    cdef isl_schedule_node *tadashi_full_shift_val(isl_schedule_node * node, long val)
    cdef isl_schedule_node *tadashi_partial_shift_val(isl_schedule_node * node, int pa_idx, long val)
    cdef isl_schedule_node *tadashi_full_shift_var(isl_schedule_node * node, long var_idx, long coeff)
    cdef isl_schedule_node *tadashi_partial_shift_var(isl_schedule_node * node, int pa_idx, long id_idx, long coeff)
    cdef isl_schedule_node *tadashi_full_shift_param(isl_schedule_node * node, long param_idx, long coeff)
    cdef isl_schedule_node *tadashi_partial_shift_param(isl_schedule_node * node, int pa_idx, long param_idx, long coeff)
    cdef isl_schedule_node *tadashi_set_parallel(isl_schedule_node * node, int num_threads)
    cdef isl_schedule_node *tadashi_set_loop_opt(isl_schedule_node * node, int pos, int opt)



#!/usb/bin/env python

# Link it to the same dir as pluto binary

import gdb.printing

import pluto_prog_printer

# pluto_prog_printer.register_printers(gdb.current_objfile())
gdb.printing.register_pretty_printer(
    gdb.current_objfile(),
    pluto_prog_printer.build_pretty_printer(),
)




f = open("results_heuristic_mt_genoa.py", "w")
f1 = open("heuristic_mt_genoa.txt")

f.write("speedups={\n")
line = f1.readline()
while line != "":
    line = line.strip()
    if line == "[STARTING NEW APP]":
        line = f1.readline().strip()
        name = line.split("/")[-1]
    if "Baseline measure" in line:
        bl = float(line.split(": ")[-1])
    if "Tiling with size 32: " in line:
        nv = float(line.split(": ")[-1])
        spd = bl/nv
        f.write("\"%s_baseline\": %f,\n"%(name, bl))
        f.write("\"%s_modified\": %f,\n"%(name, nv))
        f.write("\"%s_speedup\": %f,\n\n"%(name, spd))
    line = f1.readline()
f.write("}\n")



#!/usr/bin/env python
import re
from pathlib import Path
from subprocess import DEVNULL, PIPE, run
from typing import Optional

from colorama import Fore as CF
from tadashi import TrEnum
from tadashi.apps import App

BASE_PATH = Path(__file__).parent / "miniAMR/ref"


class miniAMR(App):
    base: Path

    def __init__(
        self,
        num_ranks: int,
        source: Path = BASE_PATH / "stencil.c",
        run_args: Optional[list[str]] = None,
        compiler_options: list = None,
        base: Path = BASE_PATH,
    ):
        self.base = base
        self.num_ranks = num_ranks
        if not run_args:
            run_args = []
        self.run_args = run_args
        include_paths = (
            self.mpich_includes()
            + self.gcc_includes("gcc")
            + self.gcc_includes("mpicc")
        )
        self._finalize_object(
            source=source,
            include_paths=include_paths,
            compiler_options=compiler_options,
        )

    @staticmethod
    def mpich_includes():
        cmd = ["mpicc", "-compile_info"]
        result = run(cmd, stdout=PIPE, stderr=DEVNULL, check=False)
        if result.returncode == 1:
            return []
        stdout = result.stdout.decode()
        opts = stdout.split()
        include_paths = [inc[2:] for inc in opts if inc.startswith("-I")]
        return include_paths

    @staticmethod
    def gcc_includes(compiler):
        cmd = [compiler, "-xc", "-E", "-v", "/dev/null"]
        result = run(cmd, stdout=DEVNULL, stderr=PIPE, check=False)
        if result.returncode == 1:
            return []
        stderr = result.stderr.decode()
        include_paths = []
        collect = False
        for line in stderr.split("\n"):
            if collect:
                if not line.startswith(" "):
                    break
                include_paths.append(line[1:])
            if line.startswith("#include <"):
                collect = True
        return include_paths

    def generate_code(self, alt_source: str = None, ephemeral=True):
        if alt_source:
            assert str(alt_source).endswith(".c")
            assert Path(alt_source).name == str(alt_source)
            new_file = self.source.parent / alt_source
        else:
            new_file = self.make_new_filename()
        self.scops.generate_code(self.source, Path(new_file))
        kwargs = {
            "source": new_file,
            "base": self.base,
            "num_ranks": self.num_ranks,
            "run_args": self.run_args,
        }
        return self.make_new_app(ephemeral, **kwargs)

    @property
    def compile_cmd(self) -> list[str]:
        cmd = [
            "make",
            "-j",
            f"-C{self.source.parent}",
            f"STENCIL={self.source.with_suffix('').name}",
            f"EXEC={self.output_binary.name}",
        ]
        return cmd

    @property
    def run_cmd(self) -> list[str]:
        cmd = ["mpirun", "-N", str(self.num_ranks), str(self.output_binary), "--stencil", "0", *self.run_args,]
        return cmd

    def extract_runtime(self, stdout: str) -> float:
        lines = stdout.split("\n")
        # scop_idx_check = set([])
        for line in lines:
            # if line.startswith("@@@") and scop_idx_check != line:
            #     scop_idx_check.add(line)
            if line.startswith("Summary:"):
                match = re.match(r".*time (\d+.\d+).*", line)
                # print(f"{scop_idx_check=}")
                return float(match.groups()[0])
        raise Exception("No output found")


def main():
    scop_idx = 0
    app = miniAMR(
        num_ranks=6,
        run_args=["--nx", "10", "--ny", "10", "--nz", "82", "--npz", "3", "--npx", "2"],
        # run_args=["--nx", "10", "--ny", "10", "--nz", "80", "--npz", "3", "--npx", "2"],
    )

    # node = app.scops[scop_idx].schedule_tree[0]
    # print(node.yaml_str)

    node = app.scops[scop_idx].schedule_tree[16]
    tr = [
                [16, TrEnum.FULL_SPLIT],
                [20, TrEnum.INTERCHANGE],
                [15, TrEnum.FULL_FUSE],
                [11, TrEnum.FULL_SPLIT],
                [10, TrEnum.FULL_FUSE],
                [6, TrEnum.FULL_SPLIT],
                [17, TrEnum.FULL_SPLIT],
                # [5, TrEnum.FULL_FUSE],
            ]
    legals = app.scops[scop_idx].transform_list(tr)
    print(f"{legals=}")
    if not all(legals):
        return
    for i, node in enumerate(app.scops[scop_idx].schedule_tree):
        at = node.available_transformations
        if at:
            print(f"{i} {at}")
    # return

    repeat=10
    app.compile()
    orig_time = app.measure(repeat)
    for ts in [61]:
        #tr = [
        #        [16, TrEnum.FULL_SPLIT],
        #        [20, TrEnum.INTERCHANGE],
        #        [15, TrEnum.FULL_FUSE],
        #        [11, TrEnum.FULL_SPLIT],
        #        [10, TrEnum.FULL_FUSE],
        #        [6, TrEnum.FULL_SPLIT],
        #        ]
        app.scops[scop_idx].reset()
        legals = app.scops[scop_idx].transform_list(tr)
        print(f"{legals=}")
        if not all(legals):
            continue
        tapp = app.generate_code(f"{ts=}.c", ephemeral=False)
        tapp.compile()
        tr_time = tapp.measure(repeat)
        speedup = orig_time / tr_time
        print(f"{ts=}")
        print(f"{orig_time=}")
        print(f"{tr_time=}")
        print(f"{speedup=}")

    print("done")


if __name__ == "__main__":
    main()



import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench

base = "examples/polybench"

gemm = Polybench(
    "linear-algebra/blas/gemm",
    base,
    compiler_options=["-DEXTRALARGE_DATASET", "-O3"],
)
print(f"{gemm.user_compiler_options=}")

gemm.compile()
print(f"{gemm.measure()=}")
for tile_size in [64, 100, 128]:
    gemm.reset_scops()
    trs = [
        [0, 2, TrEnum.FULL_SPLIT],
        [0, 3, TrEnum.TILE2D, tile_size, tile_size],
        [0, 9, TrEnum.TILE3D, tile_size, tile_size, tile_size],
    ]
    gemm.transform_list(trs)
    # print(gemm.scops[0].schedule_tree[9].yaml_str)
    tiled = gemm.generate_code(alt_infix=f"tiled{tile_size}", ephemeral=False)

    tiled.compile()
    print(f"{tile_size=} : {tiled.measure()=}")

print("DONE")



#!/bin/env python

import argparse
import logging
import random
import time
from pathlib import Path
from timeit import repeat

from tadashi import TrEnum
from tadashi.apps import Polybench, Simple

from mcts import config
from mcts.optimize import optimize_app


def main(args):
    logging.basicConfig(level=logging.INFO)
    # logger = logging.getLogger(__name__)
    # logger.info('message')
    random.seed(args.seed)
    app = Polybench(
        args.benchmark,
        compiler_options=args.compiler_options.split(" "),
    )
    print(app.scops[0].schedule_tree[0].yaml_str)
    allowed_transformations = {
        TrEnum.TILE1D,
        TrEnum.TILE2D,
        TrEnum.TILE3D,
        TrEnum.INTERCHANGE,
        # TrEnum.FUSE,
        TrEnum.FULL_FUSE,
        TrEnum.SPLIT,
        TrEnum.FULL_SPLIT,
    }
    print(f"{args.allow_omp=}")
    if args.allow_omp:
        allowed_transformations.add(TrEnum.SET_PARALLEL)

    optimize_app(
        app,
        rollouts=args.rollouts,
        repeats=args.repeats,
        whitelist_transformations=allowed_transformations,
        prefix=args.prefix,
    )
    del app
    print("all done")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("benchmark", type=str, default="stencils/jacobi-2d")
    parser.add_argument(
        "--compiler_options", type=str, default="-DEXTRALARGE_DATASET -O3"
    )
    parser.add_argument("--repeats", type=int, default=1)
    parser.add_argument("--rollouts", type=int, default=100)
    parser.add_argument("--seed", type=int, default=time.time())
    parser.add_argument(
        "--allow-omp", type=bool, default=False, action=argparse.BooleanOptionalAction
    )
    parser.add_argument("--prefix", type=str, default="data")
    args = parser.parse_args()

    main(args)



# -*- mode:cython -*-

from tadashi.ccscop cimport ccScop
from tadashi cimport isl

cdef class Scop:
  @staticmethod
  cdef Scop create(ccScop *ptr)
  cdef isl.isl_schedule_node *_cur(self)
  #
  cdef ccScop *ptr_ccscop



from tadashi import TrEnum

import mcts.base
import mcts.node_params
from mcts import config

# import MCTSNode


scop_idx = config["scop_idx"]


class MCTSNode_Transformation(mcts.base.MCTSNode):
    def set_up_children(self):
        if self.children:
            return
        node = self.app.scops[scop_idx].schedule_tree[self.action]
        available_transformations = self.filter_transformations(
            node.available_transformations
        )

        # print("AVAIL:", available_transformations)
        self.children = [
            mcts.node_params.MCTSNode_Params(parent=self, app=self.app, action=tr)
            for tr in available_transformations
        ]

    def roll(self, depth):
        self._number_of_visits += 1
        self.set_up_children()
        child = self.select_child()
        child.roll(depth)



#!/bin/env python

import abc
import copy
import datetime
import os
import re
import subprocess
import tempfile
from collections import namedtuple
from pathlib import Path
from typing import Optional

from .scop import Scop
from .translators import Pet, Translator

Result = namedtuple("Result", ["legal", "walltime"])


class App(abc.ABC):
    """The (abstract) base class for app objects."""

    user_compiler_options: list[str]
    """User compiler options are passed to the compilation command."""

    ephemeral: bool = False
    """Ephemeral, i.e. short lived apps.

    If set to `True`, the files of the `App` (usually `App.source`) is
    deleted in the destructor of the Python `App` object.

    """

    populate_scops: bool = True
    """For an `App` that is not intended for transformations.

    If set to `False` populating `App.scops` is skipped. This is
    useful to create a "shallow copy" of an `App`.

    """

    def __del__(self):
        if self.ephemeral:
            binary = self.output_binary
            if binary.exists():
                binary.unlink()
            if self.source.exists():
                self.source.unlink()
            else:
                print("WARNING: source file missing!")

    def __init__(
        self,
        source: str | Path,
        translator: Optional[Translator] = None,
        compiler_options: Optional[list[str]] = None,
        ephemeral: bool = False,
        populate_scops: bool = True,
    ):
        """Construct an app object.

        Args:

          source: The source file.
          translator: See `Translator`.
        """
        self.source = Path(source)
        if translator is None and populate_scops:
            translator = Pet()
        if compiler_options is None:
            compiler_options = []
        if populate_scops and translator:
            self.translator = translator.set_source(source, compiler_options)
        else:
            self.translator = None
        self.user_compiler_options = compiler_options
        self.ephemeral = ephemeral
        self.populate_scops = populate_scops

    def __getstate__(self):
        """This was probably needed for serialisation."""
        state = {}
        for k, v in self.__dict__.items():
            state[k] = None if k == "translator" else v
        return state

    @property
    def scops(self) -> list[Scop]:
        """The `Scop` list forwarded from `App.translator` (both for
        compatibility and convenience reasons)."""
        if self.translator == None:
            return None
        return self.translator.scops

    @property
    def legal(self) -> bool:
        return all(s.legal for s in self.scops)

    def transform_list(self, transformation_list: list) -> Result:
        for si, ni, *tr in transformation_list:
            node = self.scops[si].schedule_tree[ni]
            legal = node.transform(*tr)
        self.compile()
        walltime = self.measure()
        return Result(legal, walltime)

    def reset_scops(self):
        for scop in self.scops:
            scop.reset()

    @property
    def output_binary(self) -> Path:
        """The output binary obtained after compilation."""
        return self.source.with_suffix("")

    def generate_code(
        self,
        alt_infix=None,
        ephemeral: bool = True,
        populate_scops: bool = False,
        ensure_legality: bool = True,
    ):
        """Create a transformed copy of the app object."""
        if ensure_legality:
            if not self.scops:
                raise ValueError(
                    "The App was created without scops, cannot check legality"
                )
            if not all(s.legal for s in self.scops):
                raise ValueError("The App is not in a legal state")
        if alt_infix:
            new_file = self._source_with_infix(alt_infix)
        else:
            new_file = self._make_new_filename()
        self.translator.generate_code(str(self.source), str(new_file))
        translator = copy.copy(self.translator) if populate_scops else None
        kwargs = {
            "source": new_file,
            "translator": translator,
            "compiler_options": self.user_compiler_options,
            "populate_scops": populate_scops,
        }
        kwargs.update(self._codegen_init_args())
        app = self.__class__(**kwargs)
        app.ephemeral = ephemeral
        return app

    def _source_with_infix(self, alt_infix: str):
        mark = "INFIX"
        suffix = self.source.suffix
        pattern = rf"(.*)(-{mark}-.*)({suffix})"
        m = re.match(pattern, str(self.source))
        filename = m.groups()[0] if m else self.source.with_suffix("")
        prefix = f"{filename}-{mark}-{alt_infix}-"
        return Path(tempfile.mktemp(prefix=prefix, suffix=suffix, dir="."))

    def _make_new_filename(self) -> Path:
        mark = "TMPFILE"
        now = datetime.datetime.now()
        now_str = datetime.datetime.isoformat(now).replace(":", "-")
        suffix = self.source.suffix
        pattern = rf"(.*)(-{mark}-\d+-\d+-\d+T\d+-\d+-\d+.\d+-.*)({suffix})"
        m = re.match(pattern, str(self.source))
        filename = m.groups()[0] if m else self.source.with_suffix("")
        prefix = f"{filename}-{mark}-{now_str}-"
        return Path(tempfile.mktemp(prefix=prefix, suffix=suffix, dir="."))

    @property
    @abc.abstractmethod
    def compile_cmd(self) -> list[str]:
        """Command executed for compilation (list of strings)."""
        pass

    def _codegen_init_args(self) -> dict:
        return {}

    @staticmethod
    def compiler():
        return os.getenv("CC", "gcc")

    def compile(
        self,
        verbose: bool = False,
        extra_compiler_options: list[str] = [],
        output_binary_suffix="",
    ):
        """Compile the app so it can be measured/executed."""
        cmd = self.compile_cmd
        cmd += ["-o", f"{self.output_binary}{output_binary_suffix}"]
        cmd += self.user_compiler_options
        cmd += extra_compiler_options
        if verbose:
            print(f"{' '.join(cmd)}")
        result = subprocess.run(cmd)
        # raise an exception if it didn't compile
        result.check_returncode()

    def measure(self, repeat=1, *args, **kwargs) -> float:
        """Measure the runtime of the app."""
        if not self.output_binary.exists():
            self.compile()
        results = []
        for _ in range(repeat):
            result = subprocess.run(
                str(self.output_binary), stdout=subprocess.PIPE, *args, **kwargs
            )
            stdout = result.stdout.decode()
            results.append(self.extract_runtime(stdout))
        return min(results)

    def extract_runtime(self, stdout: str) -> float:
        """Extract the measured runtime from the output."""
        raise NotImplementedError()


class Simple(App):
    runtime_prefix: str

    def __init__(
        self,
        source: str | Path,
        translator: Optional[Translator] = None,
        compiler_options: Optional[list[str]] = None,
        ephemeral: bool = False,
        populate_scops: bool = True,
        *,
        runtime_prefix: str = "WALLTIME: ",
    ):
        self.runtime_prefix = runtime_prefix
        super().__init__(
            source,
            translator,
            compiler_options=compiler_options,
            ephemeral=ephemeral,
            populate_scops=populate_scops,
        )

    @property
    def compile_cmd(self) -> list[str]:
        return [self.compiler(), str(self.source), "-fopenmp"]

    def extract_runtime(self, stdout) -> float:
        for line in stdout.split("\n"):
            if line.startswith(self.runtime_prefix):
                num = line.split(self.runtime_prefix)[1]
                return float(num)
        return 0.0


POLYBENCH_BASE = str(Path(__file__).parent.parent / "examples/polybench")


class Polybench(App):
    """A single benchmark in of the Polybench suite."""

    benchmark: str  # path to the benchmark dir from base
    base: Path  # the dir where polybench was unpacked

    def __init__(
        self,
        benchmark: str,
        source: Optional[Path] = None,
        translator: Optional[Translator] = None,
        base: Path = Path(POLYBENCH_BASE),
        compiler_options: Optional[list[str]] = None,
        ephemeral: bool = False,
        populate_scops: bool = True,
    ):
        self.base = Path(base)
        self.benchmark = self._get_benchmark(benchmark)
        if source is None:
            filename = Path(self.benchmark).with_suffix(".c").name
            source = self.base / self.benchmark / filename
        if compiler_options is None:
            compiler_options = []
        compiler_options.append(f"-I{self.base / 'utilities'}")
        super().__init__(
            source,
            translator,
            compiler_options=compiler_options,
            ephemeral=ephemeral,
            populate_scops=populate_scops,
        )

    def _get_benchmark(self, benchmark: str) -> str:
        target = Path(benchmark).with_suffix(".c").name
        for c_file in self.base.glob("**/*.c"):
            if c_file.with_suffix(".c").name == target:
                if c_file.parent.name == "utilities":
                    break  # go to raise ValueError!
                return str(c_file.relative_to(self.base).parent)
        raise ValueError(f"Not a polybench {benchmark=}")

    def _codegen_init_args(self):
        return {
            "benchmark": self.benchmark,
            "base": self.base,
        }

    @staticmethod
    def get_benchmarks(path: str = POLYBENCH_BASE):
        benchmarks = []
        for file in Path(path).glob("**/*.c"):
            filename = file.with_suffix("").name
            dirname = file.parent.name
            if filename == dirname:
                benchmarks.append(file.parent.relative_to(path))
        return list(sorted(benchmarks))

    @property
    def compile_cmd(self) -> list[str]:
        return [
            self.compiler(),
            str(self.source),
            str(self.base / "utilities/polybench.c"),
            "-DPOLYBENCH_TIME",
            "-DPOLYBENCH_USE_RESTRICT",
            "-lm",
            "-fopenmp",
        ]

    def extract_runtime(self, stdout) -> float:
        result = 0.0
        try:
            result = float(stdout.split()[0])
        except IndexError as e:
            print(f"App probaly crashed: {e}")
        return result

    def dump_arrays(self):
        suffix = ".dump"
        self.compile(
            extra_compiler_options=["-DPOLYBENCH_DUMP_ARRAYS"],
            output_binary_suffix=suffix,
        )
        result = subprocess.run(
            f"{self.output_binary}{suffix}",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return result.stderr.decode()

    def dump_scop(self):
        src = self.source.read_text()
        lines = []
        inside_scop = False
        for line in src.split("\n"):
            if "#pragma" in line and "endscop" in line:
                break
            if inside_scop:
                print(f"{line}")
                lines.append(line)
            if "#pragma" in line and "scop" in line:
                inside_scop = True
        return "\n".join(lines)




import argparse
import time
import timeit
from pathlib import Path
from random import choice, randint, randrange, seed
from subprocess import TimeoutExpired

import multiprocess as mp
import tadashi
from tadashi import TRANSFORMATIONS, LowerUpperBound, Scops, TrEnum
from tadashi.apps import Polybench, Simple

####
####
def getDepth_aux(node, depth=0):
    cl = node.children
    if len(cl) == 0:
        return depth
    else:
        return max( [getDepth_aux(c, depth+1) for c in cl] )

def getDepth(app, node_id):
    base_node = app.scops[0].schedule_tree[node_id]
    return getDepth_aux(base_node)


def searchFor(app, tr_name):
    scops = app.scops
    ret = []
    for si in range(len(scops[0].schedule_tree)):
        s = scops[0].schedule_tree[si]
        av = s.available_transformations
        for t in av:
            if t == tr_name:
                ret.append(si)
    return ret
####
####

def random_args(node, tr):
    tiles = [TrEnum.TILE1D, TrEnum.TILE2D, TrEnum.TILE3D]
    if tr in tiles:
        tile_size = choice([2**x for x in range(5, 10)])
        return [tile_size] * (1 + tiles.index(tr))
    if tr in [TrEnum.SET_PARALLEL]:
        return [0]
    return choice(node.get_args(tr, start=-64, end=64))


def multiProcess_fitnessEval(a):
    app, trials, timeout, pre_evaluated = a

    if pre_evaluated != 0:
        return pre_evaluated

    evals = []
    for _ in range(trials):
        try:
            evals.append(app.measure(timeout=timeout))
        except TimeoutExpired:
            # If the evaluations takes too long, it gets a bad fitness
            evals.append(timeout)

    # multiplied by -1 so fitness is meant to be maximized
    return -1 * min(evals)


class Individual:
    operation_list: list = None
    fitness: float = None

    def __init__(self, op: list = []):
        self.operation_list = op

    def __str__(self):
        f = "%.8f" % self.fitness if not self.fitness is None else "Not evaluated"
        tmp = "["
        for op in self.operation_list:
            tmp += "[%s, %s, %s], " % (
                str(op[0:2]),
                "tadashi.TrEnum." + op[2].name,
                str(op[3:]),
            )
        tmp += "]"
        return "%s --- %s" % (tmp, f)

    def __gt__(self, other):
        """
        You must calculate everyones fitness before sorting
        """
        return self.getFitness() > other.getFitness()

    def generateCode(self, app_factory):
        app = app_factory.generate_code(populate_scops=True)
        app.transform_list(self.operation_list)
        tapp = app.generate_code()
        tapp.compile()
        return tapp

    def getFitness(
        self, app_factory=None, n_trials: int = None, timeout=9999, evaluations=None
    ):
        """
        app_factory and n_trials are not requires if the fitness is already calculated
        """
        if not evaluations is None and str(self.operation_list) in evaluations:
            self.fitness = evaluations[str(self.operation_list)]

        if self.fitness is None:
            try:
                app = self.generateCode(app_factory)
                evals = []
                for _ in range(n_trials):
                    try:
                        evals.append(app.measure(timeout=timeout))
                    except TimeoutExpired:
                        # If the evaluations takes too long, it gets a bad fitness
                        evals.append(timeout)
            except CalledProcessError:
                print(str(self))
                assert False

            # multiplied by -1 so fitness is meant to be maximized
            self.fitness = -1 * min(evals)

        if not evaluations is None:
            evaluations[str(self.operation_list)] = self.fitness
        return self.fitness

    def isLegal(self, app_factory=None):
        app = app_factory.generate_code(populate_scops=True)
        try:
            return app.transform_list(self.operation_list)
        except ValueError:
            # If the transform_list fails, its not a valid list
            return False

    def crossover(self, other, app_factory=None):
        # 50% chance to crossover, 0% if either parents have length 0
        if (
            randint(0, 9) < 5
            and len(self.operation_list) * len(other.operation_list) > 0
        ):
            p1 = self.operation_list
            p2 = other.operation_list

            xop1 = randint(0, len(p1) - 1)
            xop2 = randint(0, len(p2) - 1)

            o1 = p1[:xop1] + p2[:xop2]
            o2 = p2[:xop2] + p1[:xop1]

            i1 = Individual(o1)
            i2 = Individual(o2)

            ret = []
            for i in [i1, i2]:
                # print(">> XO")
                if i.isLegal(app_factory):
                    ret.append(i)
                # print("<< XO")

            return ret

        else:
            return [self, other]

    def mutate(self, app_factory=None):
        app = app_factory.generate_code(populate_scops=True)
        # 10% not mutate
        # 10% lose last operation or not mutate
        # 80% of appending a new operation at the end
        mutationType = randint(0, 9)

        # No mutation
        if mutationType == 0:
            return self

        # Delete last node, if exists:
        if mutationType == 1:
            op_list = self.operation_list[:-1]
            ret = Individual(op_list)
            return ret if ret.isLegal(app_factory) else None

        if mutationType > 1:

            found = False
            op_list = self.operation_list[:]

            app.transform_list(self.operation_list)

            scops = app.scops

            # print("Mutating 1")
            at = 0
            while not found:
                at += 1
                # print("Trying to find valid mut")
                x1 = randint(0, len(scops) - 1)
                st = scops[x1].schedule_tree
                x2 = randint(0, len(st) - 1)
                node = st[x2]

                possible = node.available_transformations
                #possible = [p for p in possible if p[1]!=TrEnum.SET_PARALLEL]
                possible = [p for p in possible if not ("TrEnum.SET_PARALLEL" in str(p[1]) or "SHIFT" in str(p[1])) ]

                if len(possible) > 0:
                    tran = possible[randint(0, len(possible) - 1)]
                    args = random_args(node, tran)

                    op = [x1, x2, tran, *args]

                    tmp_op = op_list[:]
                    tmp_op.append(op)

                    ret = Individual(tmp_op)

                    # print(">>MUT")
                    legal = ret.isLegal(app_factory)
                    # print("<<MUT")

                    found = legal
            # print("Found MUT after %d attempts" % at)

            return ret


class EvolTadashi:
    population = None
    max_gen = None
    best_individual = None
    n_trials = None
    app = None
    t_size = None
    n_threads = None
    evaluations = None

    def __init__(
        self,
        app_factory,
        population_size=30,
        max_gen=20,
        n_trials=2,
        t_size=2,
        n_threads=1,
        timeout=9999,
        use_heuristic = False,
    ):
        self.app_factory = app_factory
        self.use_heuristic = use_heuristic
        # The initial population is an individual without transformations
        # so the algorithm starts by searching for simpler solutions first

        ########
        ########
        ########

        app = self.app_factory

        full_tr_list = []

        tile_size = 32

        scops = app.scops


        trs = searchFor(app, "full_split")
        trs = [[index, TrEnum.FULL_SPLIT] for index in trs]
        trs = trs[::-1]
        for t in trs:
            scops[0].reset()
            scops[0].transform_list(full_tr_list)
            valid = scops[0].transform_list([t])
            if valid[-1]:
                full_tr_list.append(t)
        scops[0].reset()
        valid = scops[0].transform_list(full_tr_list)

        trs = searchFor(app, "tile3d")
        toRemoveFrom2D = [a for a in trs]
        toRemoveFrom2D.extend([a + 1 for a in trs])
        toRemoveFrom2D = list(set(toRemoveFrom2D))
        for t in trs:
            if t - 1 in trs:
                trs.pop(trs.index(t - 1))
        trs3D = [
            [index, TrEnum.TILE3D, tile_size, tile_size, tile_size] for index in trs[::-1]
        ]


        trs2 = searchFor(app, "tile2d")
        trs2D = [[index, TrEnum.TILE2D, tile_size, tile_size] for index in trs2[::-1]]
        trs3D.extend(trs2D)
        trs3D.sort()
        trs3D = trs3D[::-1]
        for t in trs3D:
            scops[0].reset()
            scops[0].transform_list(full_tr_list)
            valid = scops[0].transform_list([t])
            if valid[-1]:
                full_tr_list.append(t)
        scops[0].reset()

        ########
        ########
        ########


        self.population = [Individual()]

        #####
        #####

        if self.use_heuristic:
            full_tr_list = [ [0]+t for t in full_tr_list]
        else:
            full_tr_list = []
        print(full_tr_list)

        for i in range(1,len(full_tr_list)):
            ind = Individual(op=full_tr_list[:1])
            legal = ind.isLegal(self.app_factory)
            if legal:
                self.population.append( Individual(op=full_tr_list[:i]) )
            else:
                print("sus")

        #####
        #####

        self.population_size = population_size
        self.max_gen = max_gen
        self.n_trials = n_trials
        self.t_size = t_size
        self.n_threads = n_threads
        self.timeout = timeout
        self.evaluations = {}

    def tournament(self):
        """
        Requires: sorted population
        """
        return self.population[
            min([randint(0, len(self.population) - 1) for _ in range(self.t_size)])
        ]

    def fit(self):

        self.best_individual = self.population[0]
        self.best_individual.getFitness(
            self.app_factory, self.n_trials, evaluations=self.evaluations
        )
        print(
            "Measure without transformations:",
            self.best_individual.getFitness(
                self.app_factory, self.n_trials, evaluations=self.evaluations
            ),
        )

        for gen in range(self.max_gen):
            print("Gen %d" % gen)

            # print("  Calc fitness")
            # [i.getFitness( self.app_factory, self.n_trials, timeout=self.timeout ) for i in self.population]
            # self.population.sort(reverse=True)

            start_time = time.time()
            if self.n_threads > 1:
                with mp.Pool(processes=self.n_threads) as pool:
                    results = pool.map(
                        multiProcess_fitnessEval,
                        [
                            (
                                ind.generateCode(self.app_factory),
                                self.n_trials,
                                self.timeout,
                                (
                                    self.evaluations[str(ind.operation_list)]
                                    if str(ind.operation_list) in self.evaluations
                                    else 0
                                ),
                            )
                            for ind in self.population
                        ],
                    )
                    for i in range(len(self.population)):
                        self.population[i].fitness = results[i]
                        self.evaluations[str(self.population[i].operation_list)] = (
                            results[i]
                        )
            else:
                [
                    i.getFitness(
                        self.app_factory,
                        self.n_trials,
                        timeout=self.timeout,
                        evaluations=self.evaluations,
                    )
                    for i in self.population
                ]

            self.population.sort(reverse=True)

            end_time = time.time()
            print("Evaluation time:", end_time - start_time)

            print("  Fitness values obtained:")
            [print("   ", i.fitness, i) for i in self.population]

            if self.population[0] > self.best_individual:
                self.best_individual = self.population[0]
                print("  Updating best_individual to", self.best_individual)

            # print("  Breeding phase")
            new_pop = []
            while len(new_pop) < self.population_size:
                ind1 = self.tournament()
                ind2 = self.tournament()
                # print("	MUT")
                ind1 = ind1.mutate(self.app_factory)
                ind2 = ind2.mutate(self.app_factory)
                # print("	XO")
                ret = ind1.crossover(ind2, self.app_factory)
                new_pop.extend(ret)
            new_pop = new_pop[: self.population_size]

            self.population = new_pop

            print(
                "  Fitness on generation %d: %.8f (%.3fx speedup)"
                % (
                    gen,
                    self.best_individual.getFitness(self.app_factory, self.n_trials),
                    self.evaluations["[]"]/self.best_individual.getFitness(self.app_factory, self.n_trials)
                )
            )

        print("Final model:", self.best_individual)

        ########
        ########

        full_tr_list = [t[1:] for t in  self.best_individual.operation_list]

        app = self.app_factory
        scops = app.scops
        valid = scops[0].transform_list(full_tr_list)

        trs = searchFor(app, "set_parallel")
        trs = [[index, TrEnum.SET_PARALLEL, 0] for index in trs]

        trs = trs[::-1]
        tmp = []
        for t in trs:
            tmp.append( [getDepth(app, t[0]), t] )
        tmp.sort()
        trs = [tmp[-1][1]]

        for t in trs:
            scops[0].reset()
            scops[0].transform_list(full_tr_list)
            valid = scops[0].transform_list([t])
            if valid[-1]:
                full_tr_list.append(t)
        scops[0].reset()
        valid = scops[0].transform_list(full_tr_list)

        tiled = app.generate_code(alt_infix="_tiled", ephemeral=True)
        tiled.compile()

        improved = tiled.measure()
        print("Time with parallel: %f (%.3fx speedup)" % (improved, self.evaluations["[]"]/improved))


        print("FINAL transformation_list=[")
        [print("   %s," % str(t)) for t in full_tr_list]
        print("]")

        ########
        ########


def main(args):
    seed(args.seed)
    print(f"Opening {args.benchmark}")
    dataset = f"-D{args.dataset}_DATASET"
    oflag = f"-O{args.oflag}"
    print(f"Using {dataset}")
    app_factory = Polybench(args.benchmark, compiler_options=[dataset, oflag])
    app_factory.compile()
    timeout = timeit.timeit(app_factory.measure, number=1) * 2
    print("USING TIME LIMIT:", timeout)
    m = EvolTadashi(
        app_factory,
        population_size=args.population_size,
        max_gen=args.max_gen,
        n_trials=args.n_trails,
        n_threads=args.n_threads,
        use_heuristic = args.use_heuristic,
        timeout=timeout,
    )
    m.fit()


if __name__ == "__main__":
    for benchmark in ["gemm"]: #Polybench.get_benchmarks():
        print("\n\n\n")
        parser = argparse.ArgumentParser()
        parser.add_argument("--benchmark", type=str, default=benchmark)
        parser.add_argument("--dataset", type=str, default="EXTRALARGE")
        parser.add_argument("--oflag", type=int, default=3)
        parser.add_argument("--seed", type=int, default=47)
        parser.add_argument("--population-size", type=int, default=100)
        parser.add_argument("--max-gen", type=int, default=50)
        parser.add_argument("--n-trails", type=int, default=5)
        parser.add_argument("--n-threads", type=int, default=10)
        parser.add_argument("--use-heuristic", action=argparse.BooleanOptionalAction)
        #print(str(parser))
        args = parser.parse_args()
        print(str(args) + "\n")
        main(args)



#!/bin/env python
import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench


def compare(kernel, trs, rep=3):
    print(f"{kernel=}")
    compiler_options = ["-DEXTRALARGE_DATASET", "-O3"]
    app = Polybench(kernel, compiler_options=compiler_options)
    legal = app.scops[0].transform_list(trs)
    print(f"{legal=}")
    tapp = app.generate_code()

    app.compile()
    tapp.compile()
    baseline = app.measure(rep)
    transformed = tapp.measure(rep)
    print(f"{baseline=} {transformed=}")
    print(f"{baseline/transformed=}")


def main():
    gemm = [
        [1, TrEnum.FULL_SHIFT_PARAM, 2, 16],
        [
            2,
            TrEnum.FULL_FUSE,
        ],
        [1, TrEnum.SET_PARALLEL, 6],
        [2, TrEnum.FULL_SHIFT_PARAM, 1, 48],
        [2, TrEnum.TILE2D, 32, 32],
        [3, TrEnum.SET_LOOP_OPT, 0, 3],
    ]
    compare("gemm", gemm)

    jacobi2d = [
        [9, TrEnum.TILE1D, 32],
        [4, TrEnum.TILE2D, 32, 32],
        [10, TrEnum.TILE2D, 32, 32],
        [10, TrEnum.TILE1D, 32],
        [13, TrEnum.INTERCHANGE],
    ]
    compare("jacobi-2d", jacobi2d)


if __name__ == "__main__":
    main()
    print("DONE")



#!/bin/env python

import unittest
from pathlib import Path

from tadashi import apps
from tadashi.scop import TrEnum
from tadashi.translators import Pet


class TestApp(unittest.TestCase):
    examples: Path = Path(__file__).parent.parent / "examples"

    def compare_members(self, app):
        akeys = sorted(app.__dict__.keys())
        tapp = app.generate_code(ensure_legality=False)
        tkeys = sorted(tapp.__dict__.keys())
        self.assertListEqual(akeys, tkeys)
        not_equal = ["source", "ephemeral", "populate_scops", "translator"]
        for akey, aval in app.__dict__.items():
            tval = tapp.__dict__[akey]
            if akey in not_equal:
                continue
            self.assertEqual((akey, aval), (akey, tval))

    def test_app_legality(self):
        app = apps.Polybench("jacobi-1d")
        trs = [
            [0, 7, TrEnum.SET_LOOP_OPT, 0, 0],
            [0, 2, TrEnum.FULL_SPLIT],
            [0, 7, TrEnum.SET_LOOP_OPT, 0, 3],
        ]
        result = app.transform_list(trs)
        self.assertFalse(result.legal)

    def test_app_legal(self):
        app = apps.Polybench("gemm")
        trs = [
            [1, "full_shift_param", 2, 16],
            [2, "full_fuse"],
            [1, "set_parallel", 6],
            [2, "full_shift_param", 1, 48],
            [2, "tile_2d", 32, 32],
            [3, "set_loop_opt", 0, 3],
        ]
        result = app.scops[0].transform_list(trs)
        app.reset_scops()
        for legal, (ni, *args) in zip(result, trs):
            node = app.scops[0].schedule_tree[ni]
            rv = node.transform(*args)
            self.assertEqual(rv, legal)
            self.assertEqual(rv, app.legal)


class TestSimple(TestApp):
    def test_generate_code_ephemeral(self):
        """Test the `generate_code`s method `ephemeral` parameter."""
        input_file = self.examples / "inputs/depnodep.c"
        for ephemeral in [True, False]:
            expected = not ephemeral
            with self.subTest(ephemeral=ephemeral):
                app = apps.Simple(input_file, Pet())
                tapp = app.generate_code(ensure_legality=False, ephemeral=ephemeral)
                file_path = tapp.source
                del tapp
                file_exists = file_path.exists()
                if file_exists:
                    file_path.unlink()
                self.assertEqual(file_exists, expected)

    def test_args(self):
        file = self.examples / "inputs/depnodep.c"
        app = apps.Simple(file, Pet(autodetect=True))
        self.compare_members(app)


class TestPolybench(TestApp):
    base: Path = TestApp.examples / "polybench"

    def test_args(self):
        app = apps.Polybench("stencils/jacobi-2d")
        self.compare_members(app)

    def test_trlist(self):
        app = apps.Polybench("stencils/jacobi-2d")
        trs = [
            [2, TrEnum.FULL_SPLIT],
            [3, TrEnum.TILE_2D, 20, 20],
            [10, TrEnum.TILE_3D, 30, 30, 30],
        ]
        app.scops[0].transform_list(trs)
        tapp = app.generate_code(ensure_legality=False)
        tapp.compile()
        # tapp.measure()

    def test_get_benchmark(self):
        app = apps.Polybench("correlation")
        self.assertEqual(app.benchmark, "datamining/correlation")

        app = apps.Polybench("datamining/correlation")  # old
        self.assertEqual(app.benchmark, "datamining/correlation")

        with self.assertRaises(ValueError):
            app = apps.Polybench("polybench")  # special case

        with self.assertRaises(ValueError):
            app = apps.Polybench("does_not_exist")

    def test_dump_arrays(self):
        # print([a.name for a in apps.Polybench.get_benchmarks()])
        app = apps.Polybench("deriche", compiler_options=["-DMINI_DATASET"])
        # for idx, node in enumerate(app.scops[0].schedule_tree):
        #     if TrEnum.SPLIT in node.available_transformations:
        #         print(f"node[{idx}] has SPLI")
        # print(node.yaml_str)
        node = app.scops[0].schedule_tree[20]
        node.transform(TrEnum.SPLIT, 1)
        tapp = app.generate_code()
        tarrays = tapp.dump_arrays()
        tapp.measure()



from itertools import product

from tadashi import TRANSFORMATIONS, LowerUpperBound, TrEnum

import mcts.node_node
from mcts import config

from .base import MCTSNode

scop_idx = config["scop_idx"]


class MCTSNode_Params(MCTSNode):

    def get_args(self):
        node = self.app.scops[scop_idx].schedule_tree[self.parent.action]
        if self.action == TrEnum.SET_PARALLEL:
            return [[0]]
        tiles = [TrEnum.TILE1D, TrEnum.TILE2D, TrEnum.TILE3D]
        if self.action in tiles:
            rep = 1 + tiles.index(self.action)
            return [[2**x] * rep for x in range(5, 10)]
        return node.get_args(self.action, -5, 5)

    # TODO: perhaps implementing tail recursion here for var len params
    # also maybe better to make children a dictionary, so that we can add dynamically
    # TODO: this can be done lazily, if too many params
    def set_up_children(self):
        param_sets = self.get_args()
        if self.children is None:
            self.children = [
                mcts.node_node.MCTSNode_Node(parent=self, app=self.app, action=p)
                for p in param_sets
            ]

    def roll(self, depth):
        # print("select params")
        self._number_of_visits += 1
        self.set_up_children()
        child = self.select_child()
        child.evaluate()
        if depth < config["max_depth"]:
            child.roll(depth + 1)




import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench


base = "PolyBenchC-4.2.1/"
app_name = "medley/floyd-warshall"

app = Polybench(
	app_name,
	base,
	compiler_options=["-DEXTRALARGE_DATASET", "-DLARGE_DATASET", "-O3"][::2],
)

print(app_name)
print(f"{app.user_compiler_options=}")

app.compile()

#print(f"{app.measure()=}")


if False:
	scops = app.scops 
	print(len(scops[0].schedule_tree))
	for si in range(len(scops[0].schedule_tree)):
		s = scops[0].schedule_tree[si]
		av = s.available_transformations
		for t in av:
			print(si, t)


for tile_size in [32, 64, 128]: 
	app.reset_scops()

	print("TILING")

	
	trs=[ 
	[0, 1, TrEnum.TILE3D, tile_size, tile_size, tile_size],
	]

	app.transform_list(trs)

	tiled = app.generate_code(alt_infix="_tiled%d"%tile_size, ephemeral=False)
	tiled.compile()

	print("Tiling with size %d: %f" % (tile_size, tiled.measure()) )






#medley/floyd-warshall
#app.user_compiler_options=['-DEXTRALARGE_DATASET', '-O3']
#app.measure()=103.855394

# tile 3D
trs=[ 
	[0, 1, TrEnum.TILE3D, tile_size, tile_size, tile_size],
]
# tile_size=64: tiled.measure()=14.037779





import tadashi
from tadashi.apps import Polybench

base = "./examples/polybench"
app = Polybench(
    "stencils/jacobi-2d",
    base,
    compiler_options=["-O3"],
)
print(app.source)

xts = 31
yts = 31
trs = [
    [4, tadashi.TrEnum.TILE2D, xts, yts],
]
print(app.scops[0].schedule_tree[5].yaml_str)

app.scops[0].transform_list(trs)
mod = app.generate_code(f"tile{xts}-{yts}", ephemeral=False)

print(mod.source)

app.compile()
print(f"{app.measure()=}")
mod.compile()
print(f"{mod.measure()=}")




import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench


base = "PolyBenchC-4.2.1/"
app_name = "datamining/correlation"

app = Polybench(
	app_name,
	base,
	compiler_options=["-DEXTRALARGE_DATASET", "-DLARGE_DATASET", "-O3"][::2],
)

print(app_name)
print(f"{app.user_compiler_options=}")

app.compile()

print(f"{app.measure()=}")




for tile_size in [32, 64, 128, 256]: 
	app.reset_scops()

	print("TILING")

	
	trs=[ 
	[0, 41, TrEnum.FULL_SPLIT],
	[0, 36, TrEnum.FULL_SPLIT],
	[0, 29, TrEnum.FULL_SPLIT],
	[0, 14, TrEnum.FULL_SPLIT],
	[0, 4, TrEnum.FULL_SPLIT],
	[0, 53, TrEnum.TILE2D, tile_size, tile_size],
	[0, 20, TrEnum.TILE2D, tile_size, tile_size],
	[0, 8, TrEnum.TILE2D, tile_size, tile_size],
	[0, 52, TrEnum.FULL_FUSE],
	[0, 46, TrEnum.FULL_FUSE],
	[0, 38, TrEnum.FULL_FUSE],
	[0, 17, TrEnum.FULL_FUSE],
	[0, 3, TrEnum.FULL_FUSE],
	]

	app.transform_list(trs)

	if False:
		scops = app.scops 
		print(len(scops[0].schedule_tree))
		for si in range(len(scops[0].schedule_tree)):
			s = scops[0].schedule_tree[si]
			av = s.available_transformations
			for t in av:
				if "tile" in t or "split" in t or "fuse" in t:
					print(si, t)


	tiled = app.generate_code(alt_infix="_tiled%d"%tile_size, ephemeral=False)
	tiled.compile()

	print("Tiling with size %d: %f" % (tile_size, tiled.measure()) )






#datamining/correlation
#app.user_compiler_options=['-DEXTRALARGE_DATASET', '-O3']
#app.measure()=33.517485

# tile 2D
	trs=[ 
	[0, 41, TrEnum.FULL_SPLIT],
	[0, 36, TrEnum.FULL_SPLIT],
	[0, 29, TrEnum.FULL_SPLIT],
	[0, 14, TrEnum.FULL_SPLIT],
	[0, 4,  TrEnum.FULL_SPLIT],
	[0, 53, TrEnum.TILE2D, tile_size, tile_size],
	[0, 20, TrEnum.TILE2D, tile_size, tile_size],
	[0, 8,  TrEnum.TILE2D, tile_size, tile_size],
	[0, 52, TrEnum.FULL_FUSE],
	[0, 46, TrEnum.FULL_FUSE],
	[0, 38, TrEnum.FULL_FUSE],
	[0, 17, TrEnum.FULL_FUSE],
	[0, 3,  TrEnum.FULL_FUSE],
	]

# Tiling with size 32: 1.974256
# Tiling with size 64: 2.009392
# Tiling with size 128: 2.137372
# Tiling with size 256: 2.188174





# distutils: language=c++
import os
from pathlib import Path

import cython
from cython.cimports.libc.stdio import FILE, fclose, fopen
from cython.cimports.libcpp.vector import vector
from cython.cimports.tadashi import isl, pet
from cython.cimports.tadashi.ccscop import ccScop
from cython.cimports.tadashi.codegen import codegen
from cython.cimports.tadashi.scop import Scop

ABC_ERROR_MSG = "Translator is an abstract base class, use a derived class."
DOUBLE_SET_SOURCE = "Translator.set_source() should only be called once."


@cython.cclass
class Translator:
    """Translator base class."""

    ccscops = cython.declare(vector[ccScop])
    scops = cython.declare(list[Scop], visibility="public")
    ctx: isl.ctx
    source: str

    def __dealloc__(self):
        self.ccscops.clear()
        if self.ctx is not cython.NULL:
            isl.isl_ctx_free(self.ctx)

    @staticmethod
    def _check_missing_file(path: str | Path):
        path = Path(path)
        if not path.exists():
            raise ValueError(f"{path} does not exist!")

    @staticmethod
    def _get_flags(flag: str, flags: list[str]) -> list[str]:
        """Get values of the certain compiler flags from compiler options."""
        defines = []
        for i, opt in enumerate(flags):
            if not opt.startswith(f"-{flag}"):
                continue
            if opt == f"-{flag}":
                if i + 1 >= len(flags):
                    raise ValueError(f"Empty -{flag} comiler option")
                defines.append(flags[i + 1])
            else:
                defines.append(opt[2:])
        return defines

    def set_source(self, source: str | Path, options: list[str]) -> Translator:
        """Do the bookkeeping of the init.

        This method should be called from the derived class as
        super().set_source() after populatin the `ccScop`s.

        """
        self._check_missing_file(source)
        if self.ccscops.size():
            raise RuntimeError(DOUBLE_SET_SOURCE)
        self.source = str(source)
        self.scops = []
        self.ccscops.clear()
        self._populate_ccscops(str(source), options)
        for idx in range(self.ccscops.size()):
            ptr = cython.address(self.ccscops[idx])
            self.scops.append(Scop.create(ptr))
        return self

    def generate_code(self, input_path, output_path):
        raise NotImplementedError(ABC_ERROR_MSG)

    def set_include_paths(self, include_paths: list[str]):
        return self


@cython.cclass
class Pet(Translator):
    autodetect: bool

    def __init__(self, autodetect: bool = False):
        self.autodetect = autodetect

    def __copy__(self):
        cls = self.__class__
        return cls(self.autodetect)

    @cython.ccall
    def _populate_ccscops(self, source: str, options: list[str]):
        self.ctx = pet.isl_ctx_alloc_with_pet_options()
        if self.ctx is cython.NULL:
            raise MemoryError()
        # Set includes
        include_paths = self._get_flags("I", options)
        includes_str = ":".join([p for p in include_paths])
        prev_include_path = os.getenv("C_INCLUDE_PATH", "")
        os.environ["C_INCLUDE_PATH"] = includes_str
        if prev_include_path:
            os.environ["C_INCLUDE_PATH"] += f":{prev_include_path}"
        # Set defines
        for define in self._get_flags("D", options):
            pet.pet_options_append_defines(self.ctx, define.encode())
        # Set autodetect
        opt = 1 if self.autodetect else 0
        pet.pet_options_set_autodetect(self.ctx, opt)
        # Fill self.ccscops
        rv = pet.pet_transform_C_source(
            self.ctx,
            source.encode(),
            fopen("/dev/null".encode(), "w"),
            self._extract_scops_callback,
            cython.address(self.ccscops),
        )
        if -1 == rv:
            raise ValueError(
                f"Something went wrong while parsing the {source}. Is the file syntactically correct?"
            )

    @staticmethod
    @cython.cfunc
    @cython.exceptval(check=False)
    def _extract_scops_callback(
        p: isl.printer,
        scop: pet.scop,
        user: cython.p_void,
    ) -> isl.printer:
        vec = cython.cast(cython.pointer[vector[ccScop]], user)
        vec.emplace_back(scop)
        return p

    @cython.ccall
    def generate_code(self, input_path: str, output_path: str) -> int:
        r: int = 0
        scop_idx: int = 0
        output_file = cython.declare(cython.pointer[FILE])
        output_file = fopen(output_path.encode(), "w")
        scop_ptr = self.ccscops.data()
        r = pet.pet_transform_C_source(
            self.ctx,
            input_path.encode(),
            output_file,
            self._codegen_callback,
            cython.address(scop_ptr),
        )
        fclose(output_file)
        return r

    @staticmethod
    @cython.cfunc
    @cython.exceptval(check=False)
    def _codegen_callback(
        p: isl.printer,
        scop: pet.scop,
        user: cython.p_void,
    ) -> isl.printer:
        # `pptr` is a pointer to the "pointer in the outside function
        # pointing to the ccScop objects'
        pptr = cython.cast(cython.pointer[cython.pointer[ccScop]], user)
        ccscop = cython.operator.dereference(pptr)

        # todo: remove
        node = ccscop.current_node
        sn = isl.isl_schedule_node_to_str(node)
        # print(f"{sn=}")

        if not scop or not p:
            return isl.isl_printer_free(p)
        if not ccscop.modified:
            p = pet.pet_scop_print_original(scop, p)
        else:
            sched = isl.isl_schedule_node_get_schedule(ccscop.current_node)
            p = codegen(p, scop, sched)
        pet.pet_scop_free(scop)

        # increment the outer pointer.
        cython.operator.postincrement(cython.operator.dereference(pptr))
        return p



import argparse

import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench

apps_miniAMR = [
    "examples/evaluation/miniAMR/",
]


def searchFor(app, tr_name):
    scops = app.scops
    ret = []
    for si in range(len(scops[0].schedule_tree)):
        s = scops[0].schedule_tree[si]
        av = s.available_transformations
        for t in av:
            if t == tr_name:
                ret.append(si)
    return ret


def main(app_name, repeat, allow_omp):

    print("-----------------------------------------\n\n[STARTING NEW APP]")

    print(app_name)

    app = Polybench(
        app_name,
        compiler_options=["-DEXTRALARGE_DATASET", "-O3"],
    )

    print(f"{app.user_compiler_options=}")
    print(f"{repeat=}")

    app.compile()

    print("Baseline measure: %f" % app.measure(repeat=repeat))

    full_tr_list = []

    tile_size = 32

    scops = app.scops

    trs = searchFor(app, "full_split")
    trs = [[index, TrEnum.FULL_SPLIT] for index in trs]
    trs = trs[::-1]
    for t in trs:
        scops[0].reset()
        scops[0].transform_list(full_tr_list)
        valid = scops[0].transform_list([t])
        if valid[-1]:
            full_tr_list.append(t)
        else:
            print("skipped tr:", str(t))
    scops[0].reset()
    valid = scops[0].transform_list(full_tr_list)
    print("FULL_SPLIT list validity:", valid)
    # full_tr_list.extend(trs[::-1])

    trs = searchFor(app, "tile3d")
    toRemoveFrom2D = [a for a in trs]
    toRemoveFrom2D.extend([a + 1 for a in trs])
    toRemoveFrom2D = list(set(toRemoveFrom2D))
    for t in trs:
        if t - 1 in trs:
            trs.pop(trs.index(t - 1))
    trs3D = [
        [index, TrEnum.TILE3D, tile_size, tile_size, tile_size] for index in trs[::-1]
    ]
    trs2 = searchFor(app, "tile2d")
    # for t in toRemoveFrom2D:
    # 	if t in trs2:
    # 		trs2.pop(trs2.index(t))
    trs2D = [[index, TrEnum.TILE2D, tile_size, tile_size] for index in trs2[::-1]]
    trs3D.extend(trs2D)
    trs3D.sort()
    trs3D = trs3D[::-1]
    for t in trs3D:
        scops[0].reset()
        scops[0].transform_list(full_tr_list)
        valid = scops[0].transform_list([t])
        if valid[-1]:
            full_tr_list.append(t)
        else:
            print("skipped tr:", str(t))
    scops[0].reset()
    valid = scops[0].transform_list(full_tr_list)
    print("TILE 2D and 3D list validity:", valid)
    # full_tr_list.extend(trs3D[::-1])

    if allow_omp:
        trs = searchFor(app, "set_parallel")
        # trs = [[index, TrEnum.SET_PARALLEL, 0] for index in trs]
        trs = [[trs[0], TrEnum.SET_PARALLEL, 0]]
        trs = trs[::-1]
        for t in trs:
            scops[0].reset()
            scops[0].transform_list(full_tr_list)
            valid = scops[0].transform_list([t])
            if valid[-1]:
                full_tr_list.append(t)
            else:
                print("skipped tr:", str(t))
        scops[0].reset()
        valid = scops[0].transform_list(full_tr_list)
        print("SET_PARALLEL list validity:", valid)

    # trs = searchFor(app, "full_fuse")
    # trs = [ [0, index, TrEnum.FULL_FUSE] for index in trs ]
    # app.transform_list(trs[::-1])
    # full_tr_list.extend(trs[::-1])
    # full_tr_list = [ [0]+l for l in full_tr_list]

    print("transformation_list=[")
    [print("   %s," % str(t)) for t in full_tr_list]
    print("]")

    for tile_size in [32]:
        scops[0].reset()

        print("Tiling with size %d ..." % tile_size)

        valid = scops[0].transform_list(full_tr_list)
        print("Is this transformation list valid:", valid)

        tiled = app.generate_code(alt_infix="_tiled%d" % tile_size, ephemeral=False)
        tiled.compile()

        print("Tiling with size %d: %f" % (tile_size, tiled.measure(repeat=repeat)))

    print("[FINISHED APP]\n\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("benchmark", type=str, default="jacobi-2d")
    parser.add_argument("--allow-omp", action=argparse.BooleanOptionalAction)
    parser.add_argument("--repeat", type=int, default=1)
    args = parser.parse_args()
    if args.benchmark == "all":
        for benchmark in Polybench.get_benchmarks():
            main(benchmark, args.repeat, args.allow_omp)
    else:
        main(args.benchmark, args.repeat, args.allow_omp)



#!/bin/env python
import argparse
import json
from pathlib import Path

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.cm import winter as cm
from scipy.stats import gmean

plt.rcParams["text.latex.preamble"] = (
    # r"\usepackage{libertine}\usepackage{zi4}\usepackage{newtxmath}"
    r"\usepackage{newtxtext,bm}\usepackage[cmintegrals]{newtxmath}"
)
params = {
    "axes.titlesize": 16,
    "axes.labelsize": 16,
    "legend.fontsize": 16,
    "text.usetex": True,
    "font.size": 11,
    "font.family": "libertine",
    # "text.latex.unicode": True,
}
plt.rcParams.update(params)


fontsize = 12


def read_poc_output(path):
    f = open(path).readlines()
    names = []
    bl = []
    post = []
    for line_i in range(len(f)):
        if "[STARTING NEW APP]" in f[line_i]:
            if ".orig" in f[line_i + 1]:
                continue
            app_name = f[line_i + 1]
            if ": " in f[line_i + 3]:
                app_baseline = f[line_i + 3]
            else:
                app_baseline = f[line_i + 4]
            i = line_i
            while not "[FINISHED APP]" in f[i]:
                i += 1
            app_post = f[i - 1]
            # print("---------")
            # print(app_name)
            # print(app_baseline)
            # print(app_post)
            names.append(app_name.strip().split("/")[-1])
            bl.append(float(app_baseline.strip().split(": ")[1]))
            post.append(float(app_post.strip().split(": ")[1]))
    baseline = np.array(bl)
    post = np.array(post)
    names = np.array(names)
    idcs = np.argsort(names)
    df = pd.DataFrame([names[idcs], baseline[idcs], post[idcs]])
    df = df.transpose()
    df.rename(columns={1: "baseline", 2: "poc"}, inplace=True)
    df.set_index(0, inplace=True)
    df.index.name = "benchmark"
    return df


def get_ratios(baseline, post, labels):
    # Filter: keep only values where at least one is >= 0.01
    mask = ~((baseline < 0.01) & (post < 0.01))
    # mask = array1 >= 0.001

    for l in sorted(labels):
        if not l in np.array(labels)[mask]:
            print(l)

    baseline = baseline[mask]
    post = post[mask]
    labels = np.array(labels)[mask]

    ratios = baseline / post

    sorted_indices = np.argsort(labels)
    # labels = labels[sorted_indices]
    # ratios = ratios[sorted_indices]
    ratios = np.array([r if r > 1 else 1 for r in ratios])
    print(ratios)
    return ratios, labels


def get_pluto(path):
    data = {}
    for line in open(path):
        if ":::" in line:
            benchmark, rep, time = line.split(":::")
            if benchmark in data:
                old = data[benchmark]
                new = float(time)
                data[benchmark] = new if new < old else old
            else:
                data[benchmark] = float(time)

    df = pd.DataFrame(data, index=[0])
    df = df.transpose()
    df.rename(columns={0: "pluto"}, inplace=True)
    df.index.name = "benchmark"
    # print(f"{df=}")
    return df


def get_evol():
    evol_dir = Path(__file__).parent.parent / "evol"
    results = {}
    for file in evol_dir.glob("*.txt"):
        name_with_num = file.with_suffix("").name
        dash = name_with_num.rfind("-")
        name = name_with_num[:dash]
        lines = file.read_text().split("\n")
        baseline = -float(lines[3].split(":")[1])
        evol = -float(lines[-2].split("---")[1])
        results[name] = [baseline, evol]
    df = pd.DataFrame(results).transpose()
    df.rename(columns={0: "evol_baseline", 1: "evol"}, inplace=True)
    df.index.name = "benchmark"
    return df


def main(poc_path, pluto_path=None, mcts_path=None):
    data = read_poc_output(poc_path)
    if pluto_path:
        data = data.merge(get_pluto(pluto_path), on="benchmark", how="outer")
    if mcts_path:
        mcts_data = pd.read_csv(mcts_path, index_col=0)
        data = data.merge(mcts_data, on="benchmark", how="outer")
    # data = data.merge(get_evol(), on="benchmark", how="outer")
    # print(data)

    x = np.arange(len(data))
    fig, ax = plt.subplots()

    # Reference line at ratio = 1
    ax.axhline(y=1.0, color="#ff6961", linestyle="--", linewidth=2)
    ax.axhline(y=2.0, color="lightgray", linestyle="-", linewidth=1)
    ax.axhline(y=5.0, color="lightgray", linestyle="-", linewidth=1)
    ax.axhline(y=10.0, color="lightgray", linestyle="-", linewidth=1)

    poc = (data["baseline"] / data["poc"]).to_numpy().astype(np.float64)
    poc[data["baseline"] < data["poc"]] = 1  # poc doesn't do this so we do it here
    results = poc[:]
    winners = ["poc"]

    # Plot colored ratio bars
    width = 0.6 / 2
    fix = width
    kwargs = {"edgecolor": "black", "linewidth": 0.0, "zorder": 2}
    if pluto_path:
        pluto = (data["baseline"] / data["pluto"]).to_numpy().astype(np.float64)
        bars = ax.bar(
            x + 0 * width - fix,
            pluto,
            width,
            label="Pluto",
            color=cm(10),
            **kwargs,
        )
        results = np.vstack([pluto, results])
        winners = ["pluto", "poc"]
        pluto = pluto[data.index != "adi"]  # pluto failed with adi
        print(f"{gmean(pluto)=}")
    bars = ax.bar(
        x + 1 * width - fix,
        poc,
        width,
        label="Heuristic",
        color=cm(150),
        **kwargs,
    )

    print(f"{gmean(poc)=}")
    if mcts_path:
        mcts = data["mcts"].to_numpy().astype(np.float64)  # this is already speedup
        bars = ax.bar(
            x + 2 * width - fix,
            mcts,
            width,
            label="MCTS",
            color=cm(300),
            **kwargs,
        )
        results = np.vstack([results, mcts])
        winners.append("MCTS")
        print(f"{gmean(mcts)=}")

    results = np.nan_to_num(results, nan=0)
    results = np.argmax(results, axis=0)
    for i, w in enumerate(winners):
        print(f"{w}: {np.sum(results == i)}")

    # bars = ax.bar(x + 3 * width - fix, evol, width, label="EVOL", **kwargs)
    # bars = ax.bar(x+width/2, ratios, width / 2, color=bar_colors, edgecolor="black", zorder=2)
    ax.legend()

    # Labels and formatting
    ax.set_ylabel("Speedup (log scale)", fontsize=fontsize)
    ax.set_title("")
    ax.set_xticks(x)
    ax.set_xticklabels(data.index, rotation=90, fontsize=fontsize - 2)
    ax.set_yticks([0.5, 1, 2, 5, 10, 20])
    # ax.legend()

    ax.set_yscale("log")
    # Log scale (optional)
    ax.set_yticks([1, 2, 5, 10, 20])  # You can adjust the range as needed
    ax.get_yaxis().set_major_formatter(plt.ScalarFormatter())
    ax.set_ylim(ymin=0.1, ymax=350)

    # Display the plot
    plt.tight_layout()
    sm = ""
    if "_st.sh" in poc_path:
        sm = "s"
    if "_mt.sh" in poc_path:
        sm = "m"
    output_file = f"{sm}t_improvements.pdf"
    print(f"{output_file=}")
    plt.savefig(output_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("poc", type=str)
    parser.add_argument("--pluto")
    parser.add_argument("--mcts")
    args = parser.parse_args()
    main(args.poc, args.pluto, args.mcts)



#!/bin/env python

from enum import Enum, auto


class NodeType(Enum):
    """Type of the schedule tree node.

    Details: `ISL online user manual (Schedule Trees)`_.

    .. _ISL online user manual (Schedule Trees):
       https://libisl.sourceforge.io/user.html#Schedule-Trees
    """

    BAND = 0
    CONTEXT = auto()
    DOMAIN = auto()
    EXPANSION = auto()
    EXTENSION = auto()
    FILTER = auto()
    LEAF = auto()
    GUARD = auto()
    MARK = auto()
    SEQUENCE = auto()
    SET = auto()



import datetime
import json

from mcts import config


class TimestampedJsonLogger:
    """
    A logger that writes timestamped JSON lines to a file.

    The filename is generated based on the timestamp of the first
    call to the log method for a given instance of this class.
    Subsequent calls append to the same file.
    """

    def __init__(self, prefix):
        """Initializes the logger, filename is not yet set."""
        now = datetime.datetime.now()
        # Generate a filename like YYYYMMDD_HHMMSSffffff.jsonl
        # Including microseconds (%f) increases uniqueness chances
        self._filename = f"{prefix}_{now.strftime('%Y%m%d_%H%M%S')}.jsonl"
        # print(f"Initializing log file: {self._filename}") # Optional: Inform user

    def log(self, speedup, transforms, source):
        """
        Logs a value to the file.

        Creates the log file on the first call, naming it after the
        current timestamp. Appends a JSON line containing the current
        timestamp and the provided value on every call.

        Args:
            value: The data/value to log. Can be any JSON-serializable type.
        """
        current_timestamp = (
            datetime.datetime.now().isoformat()
        )  # ISO format is standard
        print(config)
        log_entry = {
            "cnt_rollouts": config["cnt_rollouts"],
            "cnt_evals": config["cnt_evals"],
            "timestamp": current_timestamp,
            "speedup": speedup,
            "transform": transforms,
            "source": source,
        }

        try:
            with open(self._filename, "a", encoding="utf-8") as f:
                json_line = json.dumps(log_entry)
                f.write(json_line + "\n")
        except IOError as e:
            print(f"Error: Could not write to log file {self._filename}: {e}")
        except TypeError as e:
            print(f"Error: Value provided is not JSON serializable: {e}")


# --- Example Usage ---

# Create an instance of the logger
# my_logger = TimestampedJsonLogger()

# print("First log call...")
# my_logger.log("System startup") # This call will create the file

# # Wait a bit (optional, just for demonstration)
# import time

# time.sleep(1)

# print("Second log call...")
# my_logger.log({"event": "UserLoggedIn", "user_id": 123})

# time.sleep(1)

# print("Third log call...")
# my_logger.log(42) # Logging a different data type

# print("Finished logging.")

# If you create another instance, it will create a *different* log file
# another_logger = TimestampedJsonLogger()
# another_logger.log("This goes to a new file")



#! /bin/env python
#
# Adopted from https://pytorch.org/tutorials/intermediate/reinforcement_q_learning.html#dqn-algorithm


# THIS FILE IS NOT FINISHED!
import argparse
import math
import random
import re
from collections import deque, namedtuple
from itertools import count, product
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import tadashi
import tadashi.apps
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import yaml

CURRENT_MARK = "CURRENT_"

torch.manual_seed(42)
random.seed(42)
np.random.seed(42)


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--num_episodes",
        default=100,
        type=int,
        help="The number episodes",
    )
    parser.add_argument(
        "--batch_size",
        default=128,
        type=int,
        help="The number of transitions sampled from the replay buffer",
    )
    parser.add_argument(
        "--gamma",
        default=0.99,
        type=float,
        help="The discount factor as mentioned in the previous section",
    )
    parser.add_argument(
        "--eps_start",
        default=0.9,
        type=float,
        help="The starting value of epsilon",
    )
    parser.add_argument(
        "--eps_end",
        default=0.05,
        type=float,
        help="The final value of epsilon",
    )
    parser.add_argument(
        "--eps_decay",
        default=1000,
        type=float,
        help="The rate of exponential decay of epsilon, higher means a slower decay",
    )
    parser.add_argument(
        "--tau",
        default=0.005,
        type=float,
        help="The update rate of the target network",
    )
    parser.add_argument(
        "--lr",
        default=1e-4,
        type=float,
        help="The learning rate of the ``AdamW`` optimizer",
    )
    parser.add_argument(
        "--embedding_size",
        default=64,
        type=int,
        help="Size of embeddings",
    )
    parser.add_argument(
        "--lstm_hidden_size",
        default=128,
        type=int,
        help="Size of hidden state vectors",
    )
    parser.add_argument(
        "--num_lstm_layers",
        default=5,
        type=int,
        help="Number of LSTM layers",
    )
    parser.add_argument(
        "--num_mcts_layers",
        default=1,
        type=int,
        help="Number of MCTS layers (1 layer = 3 sublayers)",
    )
    parser.add_argument(
        "--node_top_k",
        default=5,
        type=int,
        help="Top-k for the node (sub)layer",
    )
    parser.add_argument(
        "--tr_top_k",
        default=3,
        type=int,
        help="Top-k for the transformation (sub)layer",
    )
    parser.add_argument(
        "--args_top_k",
        default=5,
        type=int,
        help="Top-k for the args (sub)layer",
    )
    return parser.parse_args()


def get_device():
    # if GPU is to be used
    return torch.device(
        "cuda"
        if torch.cuda.is_available()
        else "mps" if torch.backends.mps.is_available() else "cpu"
    )


def tr_args(node, tr):
    if tr == tadashi.TRANSFORMATIONS[tadashi.TrEnum.TILE1D]:
        tile_size = random.choice([2**x for x in range(5, 12)])
        return [tile_size]
    tr = tadashi.TRANSFORMATIONS[tr]
    lubs = tr.available_args(node)
    args = []
    for lub in lubs:
        if isinstance(lub, tadashi.LowerUpperBound):
            lb, ub = lub
            if lb is None:
                lb = -5
            if ub is None:
                ub = 5
            tmp = list(range(lb, ub))
        else:
            tmp = [t.value for t in lub]
        args = list(product(args, tmp) if args else tmp)
    args = list(args)
    return args


def get_polybench_list():
    base = Path("examples/polybench/")
    result = []
    for p in base.glob("**"):
        if Path(p / (p.name + ".c")).exists():
            result.append(p.relative_to(base))
    return base, result


def tokenize_isl_str(isl_str: str):
    tokens = [
        r"n",
        r"ni",
        r"nj",
        r"nk",
        r"i",
        r"j",
        r"k",
        r"\d+",
        r"\+",
        r"\- ",
        r"\*",
        r"\/",
        r"=",
        r"<",
        r"<=",
        r">",
        r">=",
        r"mod",
        r"and",
        r"or",
        r'"',
        r"\(",
        r"\)",
        r"\[",
        r"\]",
        r"\}",
        r"\{",
        r"->",
        r",",
        r";",
        r":",
        r"S_\d+",
        r"L_\d+",
    ]
    pattern = re.compile("|".join([f"^{t}" for t in tokens]))

    result = []
    match = re.match(pattern, isl_str.strip())
    while isl_str and match:
        pos = match.span()[1]
        result.append(isl_str[:pos].strip())
        isl_str = isl_str[pos:].strip()
        match = re.match(pattern, isl_str.strip())
    if isl_str.strip():
        raise NotImplementedError(f"ISL string not fully parsed. Remaining: {isl_str}")
    return result


def parse_str(yaml_dict: dict, level: int) -> list:
    if yaml_dict.endswith("leaf"):
        current = yaml_dict.startswith(CURRENT_MARK)
        if current:
            yaml_dict = yaml_dict.replace(CURRENT_MARK, "")
        return [(level, current, yaml_dict)]
    return [(level, False, token) for token in tokenize_isl_str(yaml_dict)]


def traverse(yaml_dict: dict, level: int = 0) -> list:
    result = []
    if isinstance(yaml_dict, str):
        return parse_str(yaml_dict, level)
    elif isinstance(yaml_dict, list):
        for item in yaml_dict:
            result += traverse(item, level + 1)
    else:
        for k, v in yaml_dict.items():
            if k == "child":
                result += traverse(v, level + 1)
            else:
                current = False
                if k.startswith(CURRENT_MARK):
                    k = k.replace(CURRENT_MARK, "")
                    current = True
                result.append((level, current, k))
                result += traverse(v, level)
    return result


def tokenize(node: tadashi.Node, vocab: Optional[list] = None) -> dict:
    pattern = re.compile(r"# YOU ARE HERE\n *")
    yaml_str = re.sub(pattern, CURRENT_MARK, node.yaml_str)
    yaml_dict = yaml.load(yaml_str, yaml.SafeLoader)
    tokens = traverse(yaml_dict)
    if not vocab:
        vocab = ["leaf"]
        for token in tokens:
            if token[-1] not in vocab:
                vocab.append(token[-1])
    return dict(vocab=vocab, tokens=tokens)


class NodeNN(nn.Module):
    def __init__(self, scop: tadashi.Scop, args: argparse.Namespace):
        super().__init__()
        device = get_device()
        self.scop = scop
        tokenizer = tokenize(scop.schedule_tree[0])
        self.vocab = tokenizer["vocab"]
        self.rnn = nn.LSTM(
            input_size=args.embedding_size,
            hidden_size=args.lstm_hidden_size,
            num_layers=args.num_lstm_layers,
            device=device,
        )
        self.embeddings = nn.Embedding(
            len(self.vocab),
            args.embedding_size,
            device=device,
        )
        self.linear = nn.Linear(
            in_features=args.lstm_hidden_size,
            out_features=1,
            device=device,
        )

    def forward(self, node: tadashi.Node):
        device = get_device()
        tokens = tokenize(node, self.vocab)["tokens"]
        print(" ".join([t for _, _, t in tokens]))
        indices = [self.vocab.index(token) for _, _, token in tokens]
        indices_tensor = torch.LongTensor(indices).to(device)
        encoded, (h0, c0) = self.rnn(self.embeddings(indices_tensor))
        return h0[-1, :]


class NodeHeadNN(nn.Module):
    def __init__(self, args: argparse.Namespace):
        super().__init__()
        self.linear = nn.Linear(
            in_features=args.lstm_hidden_size,
            out_features=1,
            device=get_device(),
        )

    def forward(self, ht: torch.Tensor):
        return self.linear(ht)


class TranNN(nn.Module):
    def __init__(self, node_nn: NodeNN, args: argparse.Namespace):
        super().__init__()
        device = get_device()
        self.node_nn = node_nn
        in_size = args.lstm_hidden_size + len(tadashi.TrEnum)
        self.linear = nn.Linear(in_features=in_size, out_features=1, device=device)

    def forward(self, node: tadashi.Node, tr: tadashi.TrEnum):
        device = get_device()
        encoded = self.node_nn(node)
        tr_idx = torch.LongTensor([list(tadashi.TrEnum).index(tr)]).to(device)
        one_hot = nn.functional.one_hot(tr_idx, len(tadashi.TrEnum)).squeeze(0)
        combined = torch.hstack([encoded, one_hot]).to(device)
        return self.linear(combined)


class ArgsNN(nn.Module):
    def __init__(self, node_nn: NodeNN, args: argparse.Namespace):
        super().__init__()
        device = get_device()
        self.node_nn = node_nn
        max_num_args = 3
        self.max_num_args = max(
            len(tadashi.TRANSFORMATIONS[t].arg_help) for t in list(tadashi.TrEnum)
        )
        in_size = args.lstm_hidden_size + self.max_num_args
        self.linear = nn.Linear(in_features=in_size, out_features=1, device=device)

    def forward(self, node: tadashi.Node, args: list):
        device = get_device()
        encoded = self.node_nn(node)
        pad_size = self.max_num_args - len(args)
        args_tensor = torch.tensor(list(args) + [0] * pad_size).to(device)
        combined = torch.hstack([encoded, args_tensor]).to(device)
        return self.linear(combined)


def mcts(
    scops: tadashi.Scops,
    scop: tadashi.Scop,
    node_nn: NodeNN,
    node_head_nn: NodeHeadNN,
    tran_nns: dict[TranNN],
    args_nns: dict[ArgsNN],
    args: argparse.Namespace,
    app: tadashi.apps.App,
):
    for d in range(args.num_mcts_layers):
        nodes = [n for n in scop.schedule_tree if n.available_transformations]
        node_times = torch.vstack([node_nn(node) for node in nodes])
        node_top_k = torch.topk(
            input=node_head_nn(node_times),
            k=min(args.node_top_k, len(nodes)),
            dim=0,
            largest=False,
        )
        for node_idx in node_top_k.indices:
            node = nodes[node_idx]
            trs = list(node.available_transformations)
            tr_times = torch.vstack([tran_nns[tr](node, tr) for tr in trs])
            tr_top_k = torch.topk(
                torch.Tensor(tr_times),
                args.tr_top_k,
                dim=0,
                largest=False,
            )
            for tr_idx in tr_top_k.indices:
                tr = trs[tr_idx]
                trargs = list(tr_args(node, tr))
                stack = [args_nns[tr](node, ta) for ta in trargs]
                if not stack:
                    stack = [args_nns[tr](node, [])]
                args_times = torch.vstack(stack)
                times = torch.Tensor(args_times)
                k = min(args.args_top_k, times.shape[0])
                args_top_k = torch.topk(times, k, dim=0, largest=False)
                for trarg_idx in args_top_k.indices:
                    trarg = trargs[trarg_idx] if trargs else []
                    print(f">>> TRANSFORM: {tr, trarg=}")
                    legal = node.transform(tr, *trarg)
                    if not legal:
                        node.rollback()
                        t = 1000000
                    else:
                        new_app = app.generate_code()
                        new_app.compile()
                        t = new_app.measure()


def main3():
    args = get_args()
    base, results = get_polybench_list()
    # benchmark = results[29]
    benchmark = "linear-algebra/solvers/durbin"
    gemm = tadashi.apps.Polybench(benchmark=benchmark, base=base)
    scop = gemm.scops[0]
    print(scop.schedule_tree[0].yaml_str)
    node_nn = NodeNN(scop, args)
    node_head_nn = NodeHeadNN(args)
    tran_nns = dict((t, TranNN(node_nn, args)) for t in tadashi.TrEnum)
    args_nns = dict((t, ArgsNN(node_nn, args)) for t in tadashi.TrEnum)
    print(benchmark)
    mcts(
        scops=gemm.scops,
        scop=gemm.scops[0],
        node_nn=node_nn,
        node_head_nn=node_head_nn,
        tran_nns=tran_nns,
        args_nns=args_nns,
        args=args,
        app=gemm,
    )


if __name__ == "__main__":
    main3()



#!/usr/bin/env python

import unittest
from pathlib import Path

from tadashi import TrEnum
from tadashi.apps import Polybench, Simple


class TestTadashi(unittest.TestCase):
    def test_get_args(self):
        app = Simple(Path(__file__).parent.parent / "examples/inputs/shifts.c")
        node = app.scops[0].schedule_tree[3]
        tr = TrEnum.PARTIAL_SHIFT_VAR
        avail_args = node.available_args(tr)
        start, end = -3, 2
        args = node.get_args(tr, start, end)
        node.transform(tr, *args[0])
        # args have `end - start` selections for the last dim, which
        # is an infinite interval...
        self.assertEqual(len(avail_args[0]) * (end - start), len(args))

    def test_scop_legality(self):
        app = Polybench("jacobi-1d")
        trs = [
            [7, TrEnum.SET_LOOP_OPT, 0, 0],
            [2, TrEnum.FULL_SPLIT],
            [7, TrEnum.SET_LOOP_OPT, 0, 3],
        ]
        legal = app.scops[0].transform_list(trs)
        self.assertTrue(legal[0])
        self.assertFalse(legal[1])
        self.assertFalse(legal[2])



#!/usr/bin/env python
import ast
import difflib
import logging
import sys
import unittest
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench, Simple
from tadashi.translators import Pet

HEADER = "/// TRANSFORMATION: "
COMMENT = "///"


@dataclass
class TransformData:
    scop_idx: int = -1
    node_idx: int = -1
    transformation: Optional[TrEnum] = None
    transformation_args: list[int] = field(default_factory=list)


def get_inputs_path() -> Path:
    base = Path(__file__).parent.parent
    return base / "examples/inputs"


class TestCcScop(unittest.TestCase):
    @staticmethod
    def _read_app_comments(app):
        TRANSFORMATION = " TRANSFORMATION: "
        transforms = []
        target_code = []
        with open(app.source) as file:
            for commented_line in file:
                if not commented_line.startswith(COMMENT):
                    continue
                line = commented_line.strip().replace(COMMENT, "")
                if line.startswith(TRANSFORMATION):
                    transform_str = line.replace(TRANSFORMATION, "")
                    lits = ast.literal_eval(transform_str)
                    td = TransformData(*lits[:2], TrEnum(lits[2].lower()), lits[3:])
                    transforms.append(td)
                else:
                    target_line = line[1:] if line else line
                    target_code.append(target_line)
        return transforms, target_code

    def _get_generated_code(self, app: Simple):
        tapp = app.generate_code(ensure_legality=False)
        generated_code = tapp.source.read_text().split("\n")
        return [x for x in generated_code if not x.startswith(COMMENT)]

    def check(self, app_file):
        logger = logging.getLogger(self._testMethodName)
        app = Simple(app_file, Pet())
        transforms, target_code = self._read_app_comments(app)

        # transform
        logger.info("Start test")
        legality = []
        for tr in transforms:
            scop = app.scops[tr.scop_idx]  # select_scop()
            node = scop.schedule_tree[tr.node_idx]  # model.select_node(scop)
            legal = node.transform(tr.transformation, *tr.transformation_args)
            if legal is not None:
                legality.append(f"legality={legal}")

        logger.info("Transformations done")
        generated_code = self._get_generated_code(app)
        logger.info("Code generated")
        generated_code += legality
        diff = difflib.unified_diff(generated_code, target_code)
        diff_str = "\n".join(diff)
        if diff_str:
            print(f"\n{Path(__file__).parent/self._testMethodName}.c:1:1")
            print(diff_str)
        logger.info("Test finished")
        del app
        self.assertTrue(generated_code == target_code)

    @staticmethod
    def _get_node(idx):
        app = Simple("tests/dummy.c", Pet())
        node = app.scops[0].schedule_tree[idx]
        return node

    @classmethod
    def _get_band_node(cls):
        return cls._get_node(1)

    @classmethod
    def _get_sequence_node(cls):
        return cls._get_node(3)

    def test_wrong_number_of_args(self):
        node = self._get_band_node()
        self.assertRaises(ValueError, node.transform, TrEnum.TILE_1D, 2, 3)

    def test_transformation_list(self):
        app = Simple("examples/inputs/depnodep.c", Pet())
        scop = app.scops[0]
        transformations = [
            [2, tadashi.TrEnum.SET_PARALLEL, 1],
            [1, tadashi.TrEnum.SET_LOOP_OPT, 0, 3],
            [1, tadashi.TrEnum.FULL_SHIFT_VAL, -47],
            [3, tadashi.TrEnum.PARTIAL_SHIFT_VAL, 0, 39],
            [1, tadashi.TrEnum.PARTIAL_SHIFT_VAL, 0, 34],
            [3, tadashi.TrEnum.PARTIAL_SHIFT_VAR, 0, 0, -22],
            [3, tadashi.TrEnum.SET_PARALLEL, 1],
        ]
        legals = scop.transform_list(transformations)
        mod_app = app.generate_code(ensure_legality=False)
        mod_app.compile()
        for legal in legals[:-1]:
            self.assertTrue(legal)
        self.assertFalse(legals[-1])

    def test_labels(self):
        app = Polybench(Path("correlation"))
        self.assertEqual(app.scops[0].schedule_tree[27].label, "L_4")
        trs = [[27, TrEnum.TILE_1D, 11]]
        app.scops[0].transform_list(trs)
        self.assertEqual(app.scops[0].schedule_tree[27].label, "L_4-tile1d-outer")
        self.assertEqual(app.scops[0].schedule_tree[28].label, "L_4-tile1d-inner")

        app.scops[0].reset()
        trs = [[27, TrEnum.TILE_2D, 11, 13]]
        app.scops[0].transform_list(trs)
        self.assertEqual(app.scops[0].schedule_tree[27].label, "L_4-tile2d-outer")
        self.assertEqual(app.scops[0].schedule_tree[28].label, "L_5-tile2d-outer")

    def test_repeated_code_generation(self):
        base = Path(__file__).parent.parent
        app = Simple(base / "examples/inputs/simple/two_loops.c", Pet())
        node = app.scops[0].schedule_tree[1]
        node.transform(TrEnum.TILE_2D, 12, 4)
        for i in range(30):
            app = app.generate_code(populate_scops=True)

    def test_bad_deps(self):
        app = Simple("tests/bad_deps.c", Pet())
        scop = app.scops[0]
        node = scop.schedule_tree[1]
        # print(node.yaml_str)
        trs = [[2, TrEnum.FULL_SPLIT]]
        valid = scop.transform_list(trs)
        tapp = app.generate_code(ensure_legality=False)
        # print(f"{valid=}")
        # print(tapp.source.read_text())

    def test_legality_new(self):
        app = Polybench("gemm")
        self.assertTrue(app.legal)


class TestCtadashiLLVM(unittest.TestCase):
    @unittest.skip("wip llvm")
    def test_foobar(self):
        # app = Polybench("gemm")
        app = tadashi.apps.SimpleLLVM(get_inputs_path() / "depnodep.c", Pet())
        print(app.source.exists())
        # node = app.scops[0].schedule_tree[2]
        node = app.scops[0].schedule_tree[1]
        # print(node.yaml_str)
        # tr = [tadashi.TrEnum.FULL_SPLIT]
        tr = [TrEnum.TILE2D, 3, 3]
        legal = node.transform(*tr)
        # print(node.yaml_str)
        tapp = app.generate_code()
        print(f"{legal=}")
        self.assertTrue(True)


def setup():
    if "-v" in sys.argv:
        logging.basicConfig(level=logging.INFO)
    test_dir = Path(__file__).parent
    for app_path in test_dir.glob("test_*.c"):

        def ch(app_path):
            return lambda self: self.check(app_path)

        test_name = app_path.with_suffix("").name
        setattr(TestCcScop, test_name, ch(app_path))


setup()




import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench


base = "PolyBenchC-4.2.1/"
app_name = "stencils/heat-3d"

app = Polybench(
	app_name,
	base,
	compiler_options=["-DEXTRALARGE_DATASET", "-DLARGE_DATASET", "-O3"][::2],
)

print(app_name)
print(f"{app.user_compiler_options=}")

app.compile()

#print(f"{app.measure()=}")


doOnlySplit
if doOnlySplit: 
	app.reset_scops()

	print("ONLY FULL SPLIT")

	
	trs=[ 
	[0, 2, TrEnum.FULL_SPLIT],
	]

	app.transform_list(trs)

	tiled = app.generate_code(alt_infix="_onlySplit", ephemeral=False)
	tiled.compile()

	print(f"Only Split: {tiled.measure()=}")





for tile_size in [8,10,16, 32]: 
	app.reset_scops()

	trs=[ 
		[0, 2, TrEnum.FULL_SPLIT],
		[0, 4, TrEnum.TILE3D, tile_size, tile_size, tile_size],
		[0, 3, TrEnum.INTERCHANGE],
		[0, 4, TrEnum.INTERCHANGE],
		[0, 5, TrEnum.INTERCHANGE],
		[0, 13, TrEnum.TILE3D, tile_size, tile_size, tile_size],
		[0, 12, TrEnum.INTERCHANGE],
		[0, 13, TrEnum.INTERCHANGE],
		[0, 14, TrEnum.INTERCHANGE],
	]

	app.transform_list(trs)

	if False:
		scops = app.scops 
		print(len(scops[0].schedule_tree))
		for si in range(len(scops[0].schedule_tree)):
			s = scops[0].schedule_tree[si]
			av = s.available_transformations
			for t in av:
				print(si, t)

	tiled = app.generate_code(alt_infix="_tiled%d"%tile_size, ephemeral=False)
	tiled.compile()

	print(f"{tile_size=} : {tiled.measure()=}")

print("DONE")




#stencils/heat-3d
#app.user_compiler_options=['-DEXTRALARGE_DATASET', '-O3']
#app.measure()=19.512852

# Only full split
trs=[ 
	[0, 2, TrEnum.FULL_SPLIT],
]
# tiled.measure()=2.792987

# full split + tile
trs=[ 
	[0, 2, TrEnum.FULL_SPLIT],
	[0, 4, TrEnum.TILE3D, tile_size, tile_size, tile_size],
	[0, 13, TrEnum.TILE3D, tile_size, tile_size, tile_size],
]
# tile_size=8 : tiled.measure()=3.651312


# full split + tile + send n_steps to the inner loop
trs=[ 
	[0, 2, TrEnum.FULL_SPLIT],
	[0, 4, TrEnum.TILE3D, tile_size, tile_size, tile_size],
	[0, 3, TrEnum.INTERCHANGE],
	[0, 4, TrEnum.INTERCHANGE],
	[0, 5, TrEnum.INTERCHANGE],
	[0, 13, TrEnum.TILE3D, tile_size, tile_size, tile_size],
	[0, 12, TrEnum.INTERCHANGE],
	[0, 13, TrEnum.INTERCHANGE],
	[0, 14, TrEnum.INTERCHANGE],
]
# tile_size=10 : tiled.measure()=2.603448



#!/usr/bin/env python
import socket

import tadashi
from halo import Halo
from mpi4py import futures
from mpi4py.futures import MPIPoolExecutor as Executor
from tadashi.apps import App, Simple


def func(idx):
    return f"FUNC({idx}) on {socket.gethostname()}"


@Halo(spinner="dots")
def main():
    app = Simple("examples/inputs/depnodep.c")
    tlist = [
        [0, 1, tadashi.TrEnum.FULL_SHIFT_VAR, 13, 1],
    ]
    print(f"MAIN {socket.gethostname()=}")
    fs = []
    with Executor(max_workers=3) as executor:
        for idx in range(3):
            app.transform_list(tlist)
            new_app = app.generate_code(ephemeral=False)
            print(f"{new_app.source=}")
            fs.append(executor.submit(new_app.compile_and_measure))
        print(f"{len(fs)=}")
        while fs:
            done = [f for f in fs if f.done()]
            for f in done:
                print(f"{f.result()=}")
                del fs[fs.index(f)]


if __name__ == "__main__":
    main()



# -*- mode:cython -*-

from tadashi cimport isl, pet

cdef extern from "src/ccscop.h":
    cdef cppclass ccScop:
        # methods

        ccScop() except +
        ccScop(pet.scop *ps) except +

        void reset() except +
        void rollback() except +
        bint check_legality();
        # members
        isl.isl_schedule_node *current_node
        isl.isl_schedule_node *tmp_node
        bint current_legal
        bint tmp_legal
        int modified



from pathlib import Path

from tadashi import TrEnum
from tadashi.mcts.optimize import optimize_app

from app import MakefileApp

path = Path(__file__).parent / "fd3d_4.2.c"
print(f"{path.exists()=}")
app = MakefileApp(path)

print(f"cnt scops: {len(app.scops)}")

allowed_transformations = {
    TrEnum.TILE1D,
    TrEnum.TILE2D,
    TrEnum.TILE3D,
    TrEnum.INTERCHANGE,
    TrEnum.FULL_FUSE,
    # TrEnum.SPLIT,
    TrEnum.FULL_SPLIT,
}

optimize_app(
    app, rollouts=10000, scop_idx=0, whitelist_transformations=allowed_transformations
)
print("all done")



#!/usr/bin/env python
from pathlib import Path

import colorama
from tadashi import TrEnum
from tadashi.apps import Simple


def main():
    color_red = colorama.Fore.LIGHTRED_EX
    color_reset = colorama.Fore.RESET
    path = Path(__file__).parent.parent / "ecp-apps/miniAMR/openmp/stencil.c"
    print(path.exists())
    app = Simple(path)
    print(app)
    print(len(app.scops))
    node = app.scops[0].schedule_tree[0]
    # print(node.yaml_str)
    for i, node in enumerate(app.scops[0].schedule_tree):
        if node.available_transformations:
            print(f"{color_red}{i}{color_reset}: {node.available_transformations}")
    node = app.scops[0].schedule_tree[6]
    tr = [TrEnum.FULL_SPLIT]
    legal = node.transform(*tr)
    print(f"{legal=}")
    tapp = app.generate_code("foobar.c", ephemeral=False)
    print(tapp.source)
    print("DONE")


if __name__ == "__main__":
    main()



#!/bin/env python
import argparse
import json
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import gmean

plt.rcParams["text.latex.preamble"] = (
    # r"\usepackage{libertine}\usepackage{zi4}\usepackage{newtxmath}"
    r"\usepackage{newtxtext,bm}\usepackage[cmintegrals]{newtxmath}"
)
params = {
    "axes.titlesize": 16,
    "axes.labelsize": 16,
    "legend.fontsize": 16,
    "text.usetex": True,
    "font.size": 11,
    "font.family": "libertine",
    # "text.latex.unicode": True,
}
plt.rcParams.update(params)


fontsize = 12

# source https://colorkit.co/palette/c7522a-e5c185-fbf2c4-74a892-008585/
COLORS = [
    "#c7522a",
    "#e5c185",
    "#74a892",
]


def read_poc_output(path):
    f = open(path).readlines()
    names = []
    bl = []
    post = []
    for line_i in range(len(f)):
        if "[STARTING NEW APP]" in f[line_i]:
            if ".orig" in f[line_i + 1]:
                continue
            app_name = f[line_i + 1]
            if ": " in f[line_i + 3]:
                app_baseline = f[line_i + 3]
            else:
                app_baseline = f[line_i + 4]
            i = line_i
            while not "[FINISHED APP]" in f[i]:
                i += 1
            app_post = f[i - 1]
            # print("---------")
            # print(app_name)
            # print(app_baseline)
            # print(app_post)
            names.append(app_name.strip().split("/")[-1])
            bl.append(float(app_baseline.strip().split(": ")[1]))
            post.append(float(app_post.strip().split(": ")[1]))
    baseline = np.array(bl)
    post = np.array(post)
    names = np.array(names)
    idcs = np.argsort(names)
    df = pd.DataFrame([names[idcs], baseline[idcs], post[idcs]])
    df = df.transpose()
    df.rename(columns={1: "baseline", 2: "Heuristic"}, inplace=True)
    df.set_index(0, inplace=True)
    df.index.name = "benchmark"
    df["Heuristic"] = df["baseline"] / df["Heuristic"]
    print(f"{sum(df['Heuristic'].dropna() < 1)=}")
    df["Heuristic"] = np.maximum(df["Heuristic"], 1)
    return df


def get_ratios(baseline, post, labels):
    # Filter: keep only values where at least one is >= 0.01
    mask = ~((baseline < 0.01) & (post < 0.01))
    # mask = array1 >= 0.001

    for l in sorted(labels):
        if not l in np.array(labels)[mask]:
            print(l)

    baseline = baseline[mask]
    post = post[mask]
    labels = np.array(labels)[mask]

    ratios = baseline / post

    sorted_indices = np.argsort(labels)
    # labels = labels[sorted_indices]
    # ratios = ratios[sorted_indices]
    ratios = np.array([r if r > 1 else 1 for r in ratios])
    print(ratios)
    return ratios, labels


def get_pluto(path):
    data = {}
    for line in open(path):
        if ":::" in line:
            benchmark, rep, time = line.split(":::")
            if benchmark in data:
                old = data[benchmark]
                new = float(time)
                data[benchmark] = new if new < old else old
            else:
                data[benchmark] = float(time)

    df = pd.DataFrame(data, index=[0])
    df = df.transpose()
    df.rename(columns={0: "Pluto"}, inplace=True)
    df.index.name = "benchmark"
    # print(f"{df=}")
    return df


def get_evol():
    evol_dir = Path(__file__).parent.parent / "evol"
    results = {}
    for file in evol_dir.glob("*.txt"):
        name_with_num = file.with_suffix("").name
        dash = name_with_num.rfind("-")
        name = name_with_num[:dash]
        lines = file.read_text().split("\n")
        baseline = -float(lines[3].split(":")[1])
        evol = -float(lines[-2].split("---")[1])
        results[name] = [baseline, evol]
    df = pd.DataFrame(results).transpose()
    df.rename(columns={0: "evol_baseline", 1: "evol"}, inplace=True)
    df.index.name = "benchmark"
    return df


def summary(data: pd.DataFrame):
    data.drop(columns="baseline", inplace=True)
    argmax = data.to_numpy().astype(np.float64).argmax(axis=1)
    for i, k in enumerate(data.columns):
        nonan = data[k].dropna().astype(np.float64)
        best = int(sum(argmax == i))
        print(f"{best=:2};  gmean({k})={gmean(nonan)}")

    print("")


def plot(ax, data, top):
    # stats
    # results = poc[:]
    # winners = ["Heuristic"]

    # Reference line at ratio = 1
    yticks = [1, 5, 20, 80, 320]
    ax.axhline(y=1.0, color="#ff6961", linestyle="--", linewidth=0.6)
    for y in yticks:
        ax.axhline(y=y, color="lightgray", linestyle="-", linewidth=0.3)
    # Plot colored ratio bars
    x = np.arange(len(data))
    width = 0.6 / 3
    fix = width
    kwargs = {"edgecolor": "black", "linewidth": 0.0, "zorder": 2}
    for i, k in enumerate(["Pluto", "Heuristic", "MCTS"]):
        bars = ax.bar(
            x + i * width - fix,
            data[k],
            width,
            label=k,
            color=COLORS[i],
            **kwargs,
        )
        # results = np.vstack([pluto, results])
        # winners = ["pluto", "Heuristic"]
        # pluto = pluto[data.index != "adi"]  # pluto failed with adi
        # print(f"{gmean(pluto)=}")

    # results = np.nan_to_num(results, nan=0)
    # results = np.argmax(results, axis=0)
    # for i, w in enumerate(winners):
    #     print(f"{w}: {np.sum(results == i)}")

    # Labels and formatting
    ax.set_title("")
    ax.set_yticks([0.5, 1, 2, 5, 10, 20])
    if top:
        ax.set_ylabel("Single-thread speedup", fontsize=fontsize)
        ax.legend(loc="upper left")
        ax.set_xticks(x)
        ax.set_xticklabels("")
        ax.tick_params(axis="x", which="both", bottom=False, labelbottom=False)
    else:
        ax.set_ylabel("Multi-thread speedup", fontsize=fontsize)
        ax.set_xticks(x)
        ax.set_xticklabels(data.index, rotation=45, ha="right", fontsize=fontsize - 2)
    # ax.legend()

    ax.set_yscale("log")
    # Log scale (optional)
    ax.set_yticks(yticks)  # You can adjust the range as needed
    ax.get_yaxis().set_major_formatter(plt.ScalarFormatter())
    ax.set_ylim(ymin=0.8, ymax=320)


def combine(poc, pluto, mcts):
    kwargs = {"on": "benchmark", "how": "outer"}
    data = get_pluto(pluto)
    data = data.merge(read_poc_output(poc), **kwargs)
    data = data.merge(pd.read_csv(mcts, index_col=0), **kwargs)
    data["Pluto"] = data["baseline"] / data["Pluto"]
    print(f"{sum(data['Pluto'].dropna() < 1)=}")
    data["Pluto"] = np.maximum(data["Pluto"].dropna(), 1)
    return data


def main(args):
    sdata = combine(args.spoc, args.spluto, args.smcts)
    mdata = combine(args.mpoc, args.mpluto, args.mmcts)

    fig, axes = plt.subplots(nrows=2, ncols=1, figsize=(10, 5))
    ax = axes[1]

    plot(axes[0], sdata, top=True)
    summary(sdata)
    plot(axes[1], mdata, top=False)
    summary(mdata)

    # Display the plot
    plt.tight_layout()
    output_file = f"comparison.pdf"
    print(f"{output_file=}")
    plt.savefig(output_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("spoc", type=Path)
    parser.add_argument("spluto", type=Path)
    parser.add_argument("smcts", type=Path)
    parser.add_argument("mpoc", type=Path)
    parser.add_argument("mpluto", type=Path)
    parser.add_argument("mmcts", type=Path)
    args = parser.parse_args()
    main(args)



# -*- mode:cython -*-

# ---
# ctx
# ---
cdef extern from "isl/ctx.h":
    ctypedef struct isl_ctx: pass
    #
    isl_ctx* isl_ctx_alloc()
    void isl_ctx_free(isl_ctx* ctx)
ctypedef isl_ctx* ctx

# -------
# printer
# -------
cdef extern from "isl/printer.h":
    ctypedef struct isl_printer: pass
    #
    isl_printer *isl_printer_free(isl_printer *p)
ctypedef isl_printer* printer

# --------
# schedule
# --------
cdef extern from "isl/schedule.h":
    ctypedef struct isl_schedule: pass
    #
    isl_schedule *isl_schedule_free(isl_schedule *sched)
    isl_schedule_node *isl_schedule_get_root(isl_schedule *sched)
    const char *isl_schedule_to_str(isl_schedule *sched)
ctypedef isl_schedule* schedule

# ---
# set
# ---
cdef extern from "isl/set.h":
    ctypedef struct isl_set: pass
    #
    isl_set *isl_set_free(isl_set *set)
    int isl_set_dim(isl_set *set, isl_dim_type dim_type)
    const char *isl_set_get_dim_name(isl_set *set, isl_dim_type dim_type, int idx)
    ###
    ctypedef struct isl_set_list: pass
    #
    isl_set_list *isl_set_list_free(isl_set_list* sel_list)
    isl_set *isl_set_list_get_at(isl_set_list *set_list, int idx)
ctypedef isl_set* set

# ---------
# union_set
# ---------
cdef extern from "isl/union_set.h":
    ctypedef struct isl_union_set: pass
    #
    isl_union_set *isl_union_set_free(isl_union_set *uset)
    int isl_union_set_n_set(isl_union_set *uset)
    isl_set_list *isl_union_set_get_set_list(isl_union_set *uset)
ctypedef isl_union_set* union_set

# ---
# aff
# ---
cdef extern from "isl/aff.h":
    ctypedef struct isl_multi_union_pw_aff: pass
    cdef enum isl_dim_type:
            isl_dim_cst
            isl_dim_param
            isl_dim_in
            isl_dim_out
            isl_dim_set = isl_dim_out
            isl_dim_div
            isl_dim_all
    #
    const char *isl_multi_union_pw_aff_to_str(isl_multi_union_pw_aff *mupa)
    isl_multi_union_pw_aff *isl_multi_union_pw_aff_free(isl_multi_union_pw_aff *mupa)
    const char *isl_multi_union_pw_aff_get_tuple_name(isl_multi_union_pw_aff *mupa, isl_dim_type dim_type)
    int isl_multi_union_pw_aff_dim(isl_multi_union_pw_aff *mupa, isl_dim_type dim_type)
    isl_union_set *isl_multi_union_pw_aff_domain(isl_multi_union_pw_aff *mupa)
ctypedef isl_multi_union_pw_aff* multi_union_pw_aff

# -------------
# schedule_node
# -------------
cdef extern from "isl/schedule_node.h":
    ctypedef struct isl_schedule_node: pass
    cdef enum isl_schedule_node_type:
        isl_schedule_node_error = -1
        isl_schedule_node_band
        isl_schedule_node_context
        isl_schedule_node_domain
        isl_schedule_node_expansion
        isl_schedule_node_extension
        isl_schedule_node_filter
        isl_schedule_node_leaf
        isl_schedule_node_guard
        isl_schedule_node_mark
        isl_schedule_node_sequence
        isl_schedule_node_set
    #
    isl_schedule_node *isl_schedule_node_free(isl_schedule_node *node)
    isl_schedule_node *isl_schedule_node_copy(isl_schedule_node *node)
    isl_schedule_node *isl_schedule_node_root(isl_schedule_node *node)
    isl_schedule_node *isl_schedule_node_parent(isl_schedule_node *node)
    isl_schedule_node *isl_schedule_node_child(isl_schedule_node *node, int pos)
    isl_schedule_node *isl_schedule_node_first_child(isl_schedule_node *node)
    isl_schedule *isl_schedule_node_get_schedule(isl_schedule_node *node)
    const char *isl_schedule_node_to_str(isl_schedule_node *node)
    int isl_schedule_node_n_children(isl_schedule_node *node)
    isl_schedule_node_type isl_schedule_node_get_type(isl_schedule_node *node)
    isl_multi_union_pw_aff* isl_schedule_node_band_get_partial_schedule(isl_schedule_node *node)
ctypedef isl_schedule_node* schedule_node




# from .node_root import MCTSNode_Root
from timeit import default_timer as timer

import mcts.node_root
from mcts import config


def optimize_app(
    app,
    rollouts=1,
    repeats=1,
    scop_idx=0,
    max_depth=2,
    whitelist_transformations=None,
    prefix="",
):
    config["rollouts"] = rollouts
    config["repeats"] = repeats
    config["max_depth"] = max_depth
    config["scop_idx"] = scop_idx
    if whitelist_transformations:
        config["whitelist_transformations"] = whitelist_transformations
    app.compile()
    print(config)
    start_time = timer()
    initial_time = app.measure(repeat=config["repeats"])
    end_time = timer()
    total_runtime = end_time - start_time
    config["timeout"] = total_runtime * 1.5 + 1
    print("initial time:", initial_time)
    root = mcts.node_root.MCTSNode_Root(
        app=app,
        action="START",
        initial_time=initial_time,
        prefix=prefix,
    )
    for rollout in range(config["rollouts"]):
        config["cnt_rollouts"] = rollout + 1
        print(f"\n---- doing rollout {rollout}")
        root.roll()
    print("\n**************************\n")
    print("sampled tree as follows:\n")
    root.set_best()
    root.print()

    print()
    print("BEST:")
    root.print_best()
    root.show_best_source()
    del root



#!/usb/bin/env python

from pathlib import Path

from tadashi import TrEnum
from tadashi.apps import Simple


class MakefileApp(Simple):
    @property
    def compile_cmd(self) -> list[str]:
        cmd = ["make", f"-C{self.source.parent}", str(self.output_binary)]
        return cmd


def main():
    path = Path(__file__).parent / "fd3d_4.2.c"
    app = MakefileApp(path)
    tr = [[155, "interchange"], [122, "full_split"]]
    legals=app.scops[0].transform_list(tr)
    print(f"{legals=}")
    tapp = app.generate_code()
    print(f"{tapp.source=}")
    app.compile()
    tapp.compile()
    orig_time = app.measure()
    new_time = tapp.measure()
    print(f"{orig_time=}")
    print(f"{new_time=}")
    speedup = orig_time / new_time
    print(f"{speedup=}")

    print("Done")


if __name__ == "__main__":
    main()



#!/bin/env python

import unittest
from pathlib import Path
from typing import Callable

from tadashi.apps import Simple
from tadashi.node_type import NodeType
from tadashi.scop import TrEnum
from tadashi.translators import Pet


class TestScop(unittest.TestCase):
    examples: Path = Path(__file__).parent.parent / "examples"

    def _get_schedule_tree(self):
        file = self.examples / "inputs/depnodep.c"
        translator = Pet()
        translator.set_source(file, [])
        scop = translator.scops[0]
        return scop.schedule_tree

    def test_schedule_tree(self):
        """Test the basic schedule_tree functionality."""
        self.assertEqual(len(self._get_schedule_tree()), 4)


class TestNode(unittest.TestCase):
    examples: Path = Path(__file__).parent.parent / "examples"

    def _get_schedule_tree(self):
        file = self.examples / "inputs/depnodep.c"
        translator = Pet()
        translator.set_source(file, [])
        scop = translator.scops[0]
        return scop.schedule_tree

    def _assert_perperties_equal(self, target: list, fn: Callable):
        for idx, node in enumerate(self._get_schedule_tree()):
            with self.subTest(idx=idx):
                self.assertEqual(fn(node), target[idx])

    def test_node_type(self):
        target = [NodeType.DOMAIN, NodeType.BAND, NodeType.BAND, NodeType.LEAF]
        self._assert_perperties_equal(target, lambda t: t.node_type)

    def test_num_children(self):
        target = [1, 1, 1, 0]
        self._assert_perperties_equal(target, lambda t: t.num_children)

    def test_parent_idx(self):
        target = [-1, 0, 1, 2]
        self._assert_perperties_equal(target, lambda t: t.parent_idx)

    def test_index(self):
        target = [0, 1, 2, 3]
        self._assert_perperties_equal(target, lambda t: t.index)

    def test_label(self):
        target = ["", "L_0", "L_1", ""]
        self._assert_perperties_equal(target, lambda t: t.label)

    def test_location(self):
        target = [[], [0], [0, 0], [0, 0, 0]]
        self._assert_perperties_equal(target, lambda t: t.location)

    def test_loop_signature(self):
        target = [
            [],
            [{"params": ["N"], "vars": ["j", "i"]}],
            [{"params": ["N"], "vars": ["j", "i"]}],
            [],
        ]
        self._assert_perperties_equal(target, lambda t: t.loop_signature)

    def test_expr(self):
        target = [
            "",
            "[N] -> L_0[{ S_0[j, i] -> [(j)] }]",
            "[N] -> L_1[{ S_0[j, i] -> [(i)] }]",
            "",
        ]
        self._assert_perperties_equal(target, lambda t: t.expr)

    def test_children_idx(self):
        target = [[1], [2], [3], []]
        self._assert_perperties_equal(target, lambda t: t.children_idx)

    def test_yaml_str(self):
        app = Simple(self.examples / "inputs/depnodep.c", Pet())
        yamls = [
            """# YOU ARE HERE
domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
  child:
    schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
""",
            """domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  # YOU ARE HERE
  schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
  child:
    schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
""",
            """domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
  child:
    # YOU ARE HERE
    schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
""",
            """domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
  child:
    schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
    child:
      # YOU ARE HERE
      leaf
""",
        ]
        nodes = app.scops[0].schedule_tree
        for idx, node in enumerate(nodes):
            self.assertEqual(node.yaml_str, yamls[idx])

    def test_simple_transformation(self):
        first = """domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  # YOU ARE HERE
  schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
  child:
    schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
"""
        second = """domain: "[N] -> { S_0[j, i] : 0 < j < N and 0 < i < N }"
child:
  # YOU ARE HERE
  schedule: "[N] -> L_1[{ S_0[j, i] -> [(i)] }]"
  child:
    schedule: "[N] -> L_0[{ S_0[j, i] -> [(j)] }]"
"""
        app = Simple(self.examples / "inputs/depnodep.c", Pet())
        node = app.scops[0].schedule_tree[1]
        self.assertEqual(node.yaml_str, first)
        legal = node.transform(TrEnum.INTERCHANGE)
        self.assertEqual(node.yaml_str, second)



# -*- mode:cython -*-

include "isl.pxd"
cdef extern from "stdio.h":
    ctypedef struct FILE: pass

cdef extern from "pet.h":
    ctypedef struct pet_scop: pass
    isl_ctx *isl_ctx_alloc_with_pet_options()
    pet_scop *pet_scop_free(pet_scop* scop)
    int pet_options_set_autodetect(isl_ctx *ctx, int val)
    void pet_options_append_defines(isl_ctx *ctx, const char *val)
    isl_schedule *pet_scop_get_schedule(pet_scop *scop)
    int pet_transform_C_source(isl_ctx *ctx, const char *input, FILE *output, isl_printer *(*transform)(isl_printer *p, pet_scop *scop, void *user), void *user)
    isl_printer *pet_scop_print_original(pet_scop *scop, isl_printer *p);
ctypedef pet_scop* scop




import tadashi
from tadashi import TrEnum
from tadashi.apps import Polybench


base = "PolyBenchC-4.2.1/"


gemm = Polybench(
	"stencils/jacobi-2d",
	base,
	compiler_options=["-DEXTRALARGE_DATASET", "-O3"],
)

print(f"{gemm.user_compiler_options=}")

gemm.compile()

print(f"{gemm.measure()=}")

#For jacobi-2d -DEXTRALARGE, the best tile_size i found in my machine was 31
for tile_size in [5, 6, 7, 16, 64, 128]: 
	gemm.reset_scops()
	
	trs=[ 
	[0,2, TrEnum.FULL_SPLIT],
	[0, 4, TrEnum.TILE, tile_size],
	[0, 6, TrEnum.TILE, tile_size],
	[0, 3, TrEnum.INTERCHANGE],
	[0, 5, TrEnum.INTERCHANGE],
	[0, 4, TrEnum.INTERCHANGE],
	[0, 11, TrEnum.TILE, tile_size],
	[0, 13, TrEnum.TILE, tile_size],
	[0, 10, TrEnum.INTERCHANGE],
	[0, 12, TrEnum.INTERCHANGE],
	[0, 11, TrEnum.INTERCHANGE],
	]

	# Also works, not as good but has 5 operations instead of 11
	trs=[ 
	[0,2, TrEnum.FULL_SPLIT],
	[0, 4, TrEnum.TILE, tile_size],
	[0, 3, TrEnum.INTERCHANGE],
	[0, 10, TrEnum.TILE, tile_size],
	[0, 9, TrEnum.INTERCHANGE],
	]


	gemm.transform_list(trs)

	tiled = gemm.generate_code(alt_infix="_tiled%d"%tile_size, ephemeral=False)
	tiled.compile()

	print(f"{tile_size=} : {tiled.measure()=}")

print("DONE")




#!/bin/env python
import json
from pathlib import Path

import matplotlib.colors as colors
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.cm import winter as cm

# https://stackoverflow.com/questions/50976297/reduce-a-panda-dataframe-by-groups
# Direct input

plt.rcParams["text.latex.preamble"] = (
    # r"\usepackage{libertine}\usepackage{zi4}\usepackage{newtxmath}"
    r"\usepackage{newtxtext,bm}\usepackage[cmintegrals]{newtxmath}"
)
params = {
    "axes.titlesize": 16,
    "axes.labelsize": 16,
    "legend.fontsize": 16,
    "text.usetex": True,
    "font.size": 11,
    "font.family": "libertine",
    # "text.latex.unicode": True,
}
plt.rcParams.update(params)
COLORS = [
    "#74a892",
    "#fbf2c4",
    "#008585",
    "#e5c185",
    "#c7522a",
]


def get_raw_df(files):
    data = {}
    for file in files:
        name = file.with_suffix("").name
        data[name] = json.load(open(file))
    df = pd.concat({k: pd.DataFrame(v) for k, v in data.items()})
    df.index.names = ["Benchmark", "Step"]
    return df


def get_breakdown_df(files, agg_args, norm):
    df = get_raw_df(files)
    gb = df.groupby(by=["Benchmark"])
    df = gb.agg(agg_args)
    if "Total walltime" in df.columns and "Kernel walltime" in df.columns:
        df["Total walltime (diff)"] = df["Total walltime"] - df["Kernel walltime"]
        df = df.drop(["Total walltime"], axis=1)
    if norm:
        df = (df.T / df.sum(axis=1)).T
    rename = {
        "Total walltime": "Execution time",
        "Kernel walltime": "Kernel execution",
    }
    df = df.rename(columns=rename)
    df.index = df.index.str.replace("-10", "")
    df.index = df.index.str.replace("-1", "")
    return df


def breakdown(files, agg_args, norm, size):
    df = get_breakdown_df(files, agg_args, norm)
    fig, ax = plt.subplots()
    df.plot(
        kind="bar",
        stacked=True,
        ax=ax,
        color=COLORS,
        # cmap=cm,
    )
    plt.xlabel("")
    if norm:
        plt.ylabel("% of runtime")
    else:
        plt.ylabel("Runtime (sec)")
    plt.title(
        f"{size}-long transformation sequence"
        if size > 1
        else "Primitive transformations"
    )
    # plt.yscale("log")
    plt.tight_layout()
    filename = f"plot-breakdown-{'norm' if norm else 'abs'}-{size}.pdf"
    print(filename)
    plt.savefig(filename)


def breakdowns():
    agg_args = {
        # "Random transformation": "min",
        "Total walltime": "min",
        # "Kernel walltime": "min",
        "Compilation": "min",
        "Code generation": "mean",
        "Transformation + legality": "mean",
        "Extraction": "mean",
    }

    # breakdown(
    #     files=Path("./times/").glob("*-10.json"),
    #     agg_args=agg_args,
    #     norm=True,
    #     size=10,
    # )

    breakdown(
        files=Path("./times/").glob("*-10.json"),
        agg_args=agg_args,
        norm=False,
        size=10,
    )
    # breakdown(
    #     files=Path("./times/").glob("*-1.json"),
    #     agg_args=agg_args,
    #     norm=True,
    #     size=1,
    # )
    breakdown(
        files=Path("./times/").glob("*-1.json"),
        agg_args=agg_args,
        norm=False,
        size=1,
    )


def get_throughput_df(files) -> pd.DataFrame:
    df = get_raw_df(files)
    gb = df.groupby(by=["Benchmark"])
    agg_args = {
        # "Random transformation": "min",
        # "Total walltime": "min",
        # "Kernel walltime": "min",
        # "Compilation": "min",
        # "Code generation": "mean",
        "Transformation + legality": "sum",
        # "Extraction": "sum",
    }
    df = gb.agg(agg_args)
    df["Throughput"] = 10 / df["Transformation + legality"]
    df.index = df.index.str.replace("-10", "")
    df = df.drop(columns=["Transformation + legality"], axis=1)
    print(df.describe())
    return df


def throughput():
    files = Path("./times/").glob("*-10.json")
    df = get_throughput_df(files)
    fig, ax = plt.subplots()
    norm = colors.Normalize(vmin=0, vmax=df["Throughput"].max() * 1.5)
    # norm = colors.TwoSlopeNorm(vmin=0, vcenter=1.0, vmax=df["Throughput"].max())
    df.plot(
        kind="bar",
        y="Throughput",
        ax=ax,
        color=COLORS[0],  # cm(norm(df["Throughput"].to_numpy())),
    )
    plt.xlabel("")
    plt.ylabel("Iterations per second")
    plt.title("Throughput")
    ax.get_legend().remove()
    plt.tight_layout()
    filename = f"plot-throughput.pdf"
    print(filename)
    plt.savefig(filename)


if __name__ == "__main__":
    breakdowns()
    throughput()



from pathlib import Path

import mcts.node_node
from mcts.logger import TimestampedJsonLogger


class MCTSNode_Root(mcts.node_node.MCTSNode_Node):
    def __init__(self, prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        prefix_dir = Path(prefix)
        prefix_dir.mkdir(exist_ok=True)
        prefix_str = str(prefix_dir / self.app.source.name)
        self.logger = TimestampedJsonLogger(prefix_str)
        self.logger.log(1, [], self.app.source.name)
        self.speedup = 1



from tadashi import TrEnum
from tadashi.mcts.optimize import optimize_app

from app import miniAMR

app = miniAMR(run_args=["--nx", "42", "--ny", "42", "--nz", "42"])

print(f"cnt scops: {len(app.scops)}")

allowed_transformations = {
    TrEnum.TILE1D,
    TrEnum.TILE2D,
    TrEnum.TILE3D,
    TrEnum.INTERCHANGE,
    TrEnum.FULL_FUSE,
    #TrEnum.SPLIT,
    TrEnum.FULL_SPLIT,
}

optimize_app(app,
             rollouts=10000,
             scop_idx=0,
             whitelist_transformations=allowed_transformations)
print("all done")



#!/usr/bin/env python

import os

from Cython.Build import cythonize
from setuptools import Extension, setup

PROJ_ROOT = os.path.abspath(f"{__file__}/..")
PREFIX = os.path.join(PROJ_ROOT, "third_party/opt")

kwargs = {
    "libraries": ["pet", "isl"],
    "include_dirs": [os.path.join(PREFIX, "include"), "tadashi/src"],
    "library_dirs": [os.path.join(PREFIX, "lib")],
    "runtime_library_dirs": [os.path.join(PREFIX, "lib")],
    "language": "c++",
    # "undef_macros": ["NDEBUG"],
}

ext_modules = [
    Extension(
        "tadashi.scop",
        ["tadashi/scop.py", "tadashi/src/ccscop.cc", "tadashi/src/transformations.c"],
        **kwargs,
    ),
    Extension(
        "tadashi.translators",
        ["tadashi/translators.py", "tadashi/src/ccscop.cc", "tadashi/src/codegen.c"],
        **kwargs,
    ),
    Extension(
        "tadashi._tr_wrappers",
        ["tadashi/_tr_wrappers.py", "tadashi/src/transformations.c"],
        **kwargs,
    ),
]

setup(ext_modules=cythonize(ext_modules))



import logging
from subprocess import TimeoutExpired

from colorama import Fore, Style

import mcts.base
import mcts.node_transformation
from mcts import config

scop_idx = config["scop_idx"]


# This is a bit confusing, but the name implies that we are on a level where we are
# SELECTING node, maybe I should shift naming
class MCTSNode_Node(mcts.base.MCTSNode):

    def set_actions_from_nodes(self):
        nodes = self.app.scops[scop_idx].schedule_tree
        nodes_transformable = []
        for i in range(len(nodes)):
            # print(i, nodes[i])
            if self.filter_transformations(nodes[i].available_transformations):
                nodes_transformable.append(i)
                # print("trans node", i)
                # print("\t", nodes[i], nodes[i].available_transformations)
        # print("transformable nodes:", len(nodes_transformable))
        # TODO: do this lazily to avoid expensive cloning through code generation
        self.children = [
            mcts.node_transformation.MCTSNode_Transformation(
                parent=self, app=self.app, action=node
            )
            for node in nodes_transformable
        ]

    def get_transform_chain(self):
        # print("GETING CHAIN from", self)
        # self.print()
        # print("-----")
        if self.parent.parent.parent.parent:
            # print(self.parent.parent.parent.parent)
            transforms = self.parent.parent.parent.get_transform_chain()
        else:
            transforms = []
        node = self.parent.parent.action
        tr = self.parent.action
        args = self.action
        transforms.append([node, tr, *args])
        return transforms

    def evaluate(self):
        self._number_of_visits += 1
        if self._number_of_visits > 4:
            return
        trs = self.get_transform_chain()
        print("selected transform:", trs)
        self.app.reset_scops()
        try:
            legal = self.app.scops[scop_idx].transform_list(trs)[-1]
            print("transform legal: ", legal)
            if legal:
                config["cnt_evals"] += 1
                # print("try generate code")
                new_app = self.app.generate_code(ephemeral=False)
                # print("try compile")
                new_app.compile()
                new_time = new_app.measure(
                    repeat=config["repeats"], timeout=config["timeout"]
                )
                print("optimized time:", new_time)
                speedup = self.get_initial_time() / new_time
                # self.source = new_app.source
                self.update_stats(speedup, trs, new_app.source.name)
            else:
                speedup = -1
                self.update_stats(speedup, trs, "")
            print("speedup:", speedup)
        except TimeoutExpired:
            print("timed out")
            self.update_stats(-1, trs, "")
        except Exception as e:
            print(Fore.RED, end="")
            print("failed to transform with the following exception:")
            print(e)
            print(Style.RESET_ALL, end="")
        finally:
            self.app.reset_scops()

    def roll(self, depth=1):
        logging.info("selecting a node to transform")
        # print("replaying transforms up to current")
        if self.parent:
            self.app.reset_scops()
            trs = self.get_transform_chain()
            self.app.scops[scop_idx].transform_list(trs)
        self._number_of_visits += 1
        if self.children is None:
            self.set_actions_from_nodes()
        # print("selected node as ", child)
        if self.children:
            child = self.select_child()
            child.roll(depth)
        else:
            print("NO NODES TO CHOOSE FROM")
            self.update_stats(-1)
